import json
import os
import sys
import tempfile

# To ensure the unit test can be run from any point in the filesystem,
# add parent folder to path to permit relative imports
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import feattranslator
from query_features import ALL_FEATURES_FILE
from features import Cmd, Endpoint, Change, Task, Ensure, Interface, Status


class TestMongoUpload:

    def test_all_features_translation(self):
        data = {
            "changes": ["pre-download"],
            "commands": ["abort", "debug execution apparmor"],
            "endpoints": [
                {"method": "GET", "path": "/v2/system-info"},
                {
                    "actions": ["advise-system-key-mismatch"],
                    "method": "POST",
                    "path": "/v2/system-info",
                },
                {
                    "actions": ["connect", "disconnect"],
                    "method": "POST",
                    "path": "/v2/interfaces",
                },
            ],
            "ensures": [
                {"function": "ensureAliasesV2", "manager": "SnapManager"},
                {
                    "function": "autoRefresh.ensureLastRefreshAnchor",
                    "manager": "SnapManager",
                },
            ],
            "interfaces": ["accel"],
            "tasks": [
                {"kind": "hotplug-add-slot"},
                {"has-undo": True, "kind": "migrate-snap-home"},
            ],
        }
        with tempfile.TemporaryDirectory() as tmpdir:
            with open(
                os.path.join(tmpdir, ALL_FEATURES_FILE), "w", encoding="utf-8"
            ) as f:
                json.dump(data, f)
            feattranslator.translate_all_features(
                os.path.join(tmpdir, ALL_FEATURES_FILE),
                os.path.join(tmpdir, "my-output.json"),
            )
            with open(
                os.path.join(tmpdir, "my-output.json"), "r", encoding="utf-8"
            ) as f:
                rewritten = json.load(f)
                assert "changes" in rewritten
                assert len(rewritten["changes"]) == 1
                assert rewritten["changes"][0] == Change(kind=data["changes"][0])
                assert "cmds" in rewritten
                assert len(rewritten["cmds"]) == 2
                assert Cmd(cmd=data["commands"][0]) in rewritten["cmds"]
                assert Cmd(cmd=data["commands"][1]) in rewritten["cmds"]
                assert "endpoints" in rewritten
                assert len(rewritten["endpoints"]) == 4
                for ep in data["endpoints"]:
                    if "actions" in ep:
                        for act in ep["actions"]:
                            assert (
                                Endpoint(
                                    method=ep["method"], path=ep["path"], action=act
                                )
                                in rewritten["endpoints"]
                            )
                    else:
                        assert (
                            Endpoint(method=ep["method"], path=ep["path"])
                            in rewritten["endpoints"]
                        )
                assert "ensures" in rewritten
                assert len(rewritten["ensures"]) == 2
                for en in data["ensures"]:
                    assert (
                        Ensure(manager=en["manager"], function=en["function"])
                        in rewritten["ensures"]
                    )
                assert "interfaces" in rewritten
                assert len(rewritten["interfaces"]) == 1
                assert Interface(name=data["interfaces"][0]) in rewritten["interfaces"]
                assert "tasks" in rewritten
                assert len(rewritten["tasks"]) == 5
                for status in [Status.done, Status.error]:
                    assert (
                        Task(kind=data["tasks"][0]["kind"], last_status=status)
                        in rewritten["tasks"]
                    )
                for status in [Status.done, Status.error, Status.undone]:
                    assert (
                        Task(kind=data["tasks"][1]["kind"], last_status=status)
                        in rewritten["tasks"]
                    )



#!/usr/bin/env python3

import argparse
from collections import namedtuple
import json
import os
import shutil
from typing import Any

import features


SpreadTaskNames = namedtuple(
    'SpreadTaskNames', ['original', 'suite', 'task', 'variant'])


def _remove_json_extension(file_name: str) -> str:
    return os.path.splitext(file_name)[0] if file_name.endswith('.json') else file_name


def _parse_file_name(file_name: str) -> SpreadTaskNames:
    '''
    Given a file name in the format with double slashes <backend>:<system>:suite--path--task:variant
    and optionally a json extension, it returns the original name, the suite name, the task name, 
    and the variant name. So in the example, it returns:
    - original_name = <backend>:<system>:suite/path/task:variant
    - suite_name = suite/path
    - task_name = task
    - variant_name = variant

    :param file_name: The file name to parse
    :returns: A namedtuple with the original name, the suite name, the task name and the variant name. If variant is not present, it's value is an empty string.
    '''
    file_name = _remove_json_extension(file_name)
    original_name = file_name.replace('--', '/')
    task = ':'.join(original_name.split(':')[2:])
    suite_name = '/'.join(task.split('/')[:-1])
    task_name = task.split('/')[-1]
    variant_name = ''
    if task_name.count(':') == 1:
        variant_name = task_name.split(':')[1]
        task_name = task_name.split(':')[0]
    return SpreadTaskNames(original_name, suite_name, task_name, variant_name)


def _compose_test(dir: str, file: str, failed_tests: set[str]) -> features.TaskFeatures:
    '''
    Creates a dictionary with the features of a test and test information.
    The features are read from the file and the test information is extracted from the file name.

    :param dir: The directory where the file is located
    :param file: The file name
    :param failed_tests: String containing the names of failing tests
    :returns: A dictionary with test information and features
    '''
    with open(os.path.join(dir, file), 'r', encoding='utf-8') as f:
        original, suite_name, result_name, variant_name = _parse_file_name(
            file)
        task_features = features.TaskFeatures(
            suite=suite_name,
            task_name=result_name,
            variant=variant_name,
            success=original not in failed_tests
        )
        task_features.update(json.loads(f.read()))
        return task_features


def _compose_env_variables(env_variables: list[str]) -> list[features.EnvVariables]:
    '''
    Given environment variables as a list of strings key=value, it creates
    a list of dictionaries of [{"name": <env1-name>, "value": <env1-value>}...]

    :param env_variables: a list of strings with key=value environment variables
    :returns: A list of dictionaries
    '''
    composed = []
    for env in env_variables:
        name, sep, value = env.partition('=')
        if sep != '=':
            raise ValueError("Not a key=value pair {}".format(env))
        composed.append(features.EnvVariables(
            name=name.strip(), value=value.strip()))
    return composed


def compose_system(dir: str, system: str, failed_tests: set[str], env_variables: list[str], scenarios: list[str]) -> features.SystemFeatures:
    '''
    Given a containing directory, a system-identifying string, and other information
    about failed tests, environment variables, and scenarios, it creates a dictionary 
    containing the feature information found in the files contained in the directory 
    for that system.

    :param dir: Directory that contains feature-tagging files
    :param system: Identifying string to select only files with that string
    :param failed_tests: String containing the names of failing tests
    :param env_variables: List of strings with key=value environment variables
    :param scenarios: List of strings with scenario names
    :returns: Dictionary containing all tests and tests information for the system
    '''
    files = [file for file in os.listdir(
        dir) if system in file and file.count(':') >= 2]
    return features.SystemFeatures(
        schema_version='0.0.0',
        system=system,
        scenarios=[scenario.strip()
                   for scenario in scenarios] if scenarios else [],
        env_variables=_compose_env_variables(env_variables),
        tests=[_compose_test(dir, file, failed_tests) for file in files]
    )


def get_system_list(dir: str) -> set[str]:
    '''
    Constructs a list of all systems from the filenames in the specified directory

    :param dir: Directory containing feature-tagging information for tests
    :returns: Set of identifying strings for systems
    '''
    files = [f for f in os.listdir(dir)
             if os.path.isfile(os.path.join(dir, f))]
    return {':'.join(file.split(':')[:2])
            for file in files if file.count(':') >= 2}


def _replace_tests(old_json_file: str, new_json_file: str) -> features.SystemFeatures:
    '''
    The new_json_file contains a set of the tests, some of which may be found
    in the old_json_file, some of which may not.
    This function leaves not-rerun tests untouched, while replacing old test
    runs with their rerun counterparts found in new_json_file. The resulting
    json in output therefore contains a mix of tests that were not rerun and
    the latest version of tests that were rerun.

    :param old_json_file: file path of first run of composed features
    :param new_json_file: file path of rerun of composed features
    :returns: dictionary that contains the first run data with rerun tests 
    replaced by the rerun data from the new_json_file
    '''
    with open(old_json_file, 'r', encoding='utf-8') as f:
        old_json = json.load(f)
    with open(new_json_file, 'r', encoding='utf-8') as f:
        new_json = json.load(f)
    for test in new_json['tests']:
        found = False
        for old_test in old_json['tests']:
            if old_test['task_name'] == test['task_name'] and old_test['suite'] == test['suite'] and old_test['variant'] == test['variant']:
                found = True
                old_test.clear()
                for key, value in test.items():
                    old_test[key] = value
                break
        if not found:
            old_json['tests'].append(test)
    return old_json


def _get_original_and_rerun_list(filenames: list[str]) -> tuple[list[str], list[str]]:
    '''
    Given a list of filenames, gets two lists of rerun information: 
    the first list contains the first run (of systems that were rerun) 
    while the second list contains all reruns, sorted from earliest to latest.

    Note: the list of first runs ONLY contains the first run of reruns;
    it does not contain systems that had no rerun.

    :param filenames: a list of filenames
    :returns: the list of first runs and the list of all reruns
    '''
    reruns = [file for file in filenames if not _remove_json_extension(
        file).endswith('_1')]
    originals = [file for file in filenames
                 if _remove_json_extension(file).endswith('_1') and
                 any(rerun for rerun in reruns if rerun.startswith(_remove_json_extension(file)[:-2]))]
    reruns.sort(key=lambda x: int(_remove_json_extension(x).split('_')[-1]))

    # If something went drastically wrong during testing, there might not be an original run.
    # ( e.g. we have my-system_2.json and my-system_3.json but not my-system_1.json)
    # Search for any reruns that don't have an original and remove them from the rerun list.
    # Then, if there is another rerun present, add that instance to the originals.
    no_originals = []
    has_rerun = set()
    for rerun in reruns:
        bare_filename = _get_name_without_run_number(rerun)
        if any(o for o in no_originals if o.startswith(bare_filename)):
            has_rerun.add(bare_filename)
            continue
        if not any(original for original in originals if original.startswith(bare_filename)):
            no_originals.append(rerun)
    for no_original in no_originals:
        reruns.remove(no_original)
        if _get_name_without_run_number(no_original) in has_rerun:
            originals.append(no_original)

    return originals, reruns


def _get_name_without_run_number(test: str) -> str:
    '''
    Given a name like <some-name>_<some-number> (optionally with extension), 
    returns <some-name>. If the name doesn't end with _<some-number>, then 
    it will return the original name without extension.
    '''
    test_split = _remove_json_extension(test).split('_')
    if test_split[-1].isdigit():
        return '_'.join(test_split[:-1])
    return _remove_json_extension(test)


def replace_old_runs(dir: str, output_dir: str) -> None:
    '''
    Given the directory in input (dir) that contains a set of files of original
    run data together with rerun data, this populates the specified output_dir
    with a consolidated set of composed features, one per system. An original
    composed features file is a file that ends in _1.json. A rerun composed
    features file is a file that ends in _<num>.json where <num> is greater 
    than 1. The numbering is automatically generated when the compose features
    script was called with the --run-attempt


    :param dir: directory containing composed feature files with varying run 
    attempt numbers
    :param output_dir: directory where to write the consolidated composed features
    '''
    filenames = [f for f in os.listdir(
        dir) if os.path.isfile(os.path.join(dir, f))]
    originals, reruns = _get_original_and_rerun_list(filenames)
    for rerun in reruns:
        result_name = _get_name_without_run_number(rerun)
        original = list(
            filter(lambda x: x.startswith(result_name), originals))
        if len(original) != 1:
            raise RuntimeError(
                f'The rerun {rerun} does not have a corresponding original run')
        original_file=os.path.join(dir, original[0])
        result_file=os.path.join(output_dir, result_name + '.json')
        if os.path.isfile(result_file):
            original_file = result_file
        tests = _replace_tests(original_file, os.path.join(dir, rerun))
        with open(result_file, 'w', encoding='utf-8') as f:
            f.write(json.dumps(tests))

    # Search for system test results that had no reruns and
    # simply copy their result file to the output folder
    for file in filenames:
        if file not in originals and file not in reruns:
            shutil.copyfile(os.path.join(dir, file),
                            os.path.join(output_dir, _get_name_without_run_number(file) + '.json'))


def main():
    description = '''
    Can be run in two modes: composed feature generation or composed feature consolidation

    Composed feature generation mode

    Given a directory containing files with outputs of journal-analzyer.py with filenames
    of format <backend>:<system>:suite--path--<test>:<variant>, it will construct a json
    file for each <backend>:<system> with feature-tagging information, accompanied with
    additional test information.

    Composed feature consolidation mode

    Given a directory containing files of pre-composed feature information with filenames like
    <backend>:<system>_<run-attempt>.json, it writes the consolidated feature information in a
    new directory (specified with the --output flag) where the latest rerun data replaces the old.
    So if a file contains one test that was later rerun, the new consolidated file will contain
    unaltered content from the original run except for the one test rerun that will replace
    the old.
    '''
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument('-d', '--dir', type=str, required=True,
                        help='Path to the folder containing json files')
    parser.add_argument('-o', '--output', type=str,
                        help='Output directory', required=True)
    parser.add_argument('-s', '--scenarios', type=str, nargs='*',
                        help='List of useful metadata tags to describe the testing scenario', default='')
    parser.add_argument('-e', '--env-variables', type=str, nargs='*',
                        help='List of environment variables as key=value', default='')
    parser.add_argument('-f', '--failed-tests', type=argparse.FileType('r'),
                        help='File containing the space-separated names of failed tests')
    parser.add_argument('--run-attempt', type=int, choices=range(1, 10), help='''
                        Run attempt number of the json files contained in the folder [1,10). 
                        Only needed when rerunning spread for failed tests. When specified, will append the run attempt 
                        number on the filename, which will then be used when running this script with the --replace-old-runs
                        flag to determine replacement order''')
    parser.add_argument('-r', '--replace-old-runs', action='store_true',
                        help='When set, will process pre-composed runs and consolidate them into the output dir')
    args = parser.parse_args()

    os.makedirs(args.output, exist_ok=True)

    if args.replace_old_runs:
        replace_old_runs(args.dir, args.output)
        exit(0)

    failed_tests = set()
    if args.failed_tests:
        for failed_test in args.failed_tests:
            failed_tests.update(failed_test.split())

    attempt = ''
    if args.run_attempt:
        attempt = '_%s' % args.run_attempt
    systems = get_system_list(args.dir)
    for system in systems:
        composed = compose_system(dir=args.dir, 
                                  system=system,
                                  failed_tests=failed_tests,
                                  env_variables=args.env_variables, 
                                  scenarios=args.scenarios)
        with open(os.path.join(args.output, system + attempt + '.json'), 'w', encoding='utf-8') as f:
            json.dump(composed, f)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from abc import abstractmethod, ABC
import argparse
from collections import defaultdict
import concurrent.futures
from contextlib import closing
import copy
import datetime
import json
import os
import pymongo
import pymongo.collection
import sys
from typing import Any, Iterable

from features import SystemFeatures, Cmd, Endpoint, Status, Task, Change, Interface


KNOWN_FEATURES = ['cmds', 'endpoints',
                  'ensures', 'tasks', 'changes', 'interfaces']

# file name for the complete list of all features
ALL_FEATURES_FILE = 'all-features.json'


class TaskId:
    suite: str
    task_name: str

    def __init__(self, suite: str, task_name: str) -> None:
        self.suite = suite
        self.task_name = task_name

    def __eq__(self, value) -> bool:
        if not isinstance(value, TaskId):
            return False
        return self.suite == value.suite and self.task_name == value.task_name

    def __hash__(self) -> int:
        return hash((self.suite, self.task_name))

    def __str__(self) -> str:
        return self.suite + "/" + self.task_name


class TaskIdVariant(TaskId):
    variant: str

    def __init__(self, suite: str, task_name: str, variant: str) -> None:
        super().__init__(suite, task_name)
        self.variant = variant

    def __eq__(self, value) -> bool:
        if isinstance(value, TaskIdVariant):
            return self.suite == value.suite and self.task_name == value.task_name and self.variant == value.variant
        elif isinstance(value, TaskId):
            return self.suite == value.suite and self.task_name == value.task_name
        else:
            return False

    def __hash__(self) -> int:
        return hash((self.suite, self.task_name, self.variant))

    def __str__(self) -> str:
        if self.variant:
            return self.suite + ":" + self.task_name + ":" + self.variant
        return self.suite + ":" + self.task_name


class DateTimeEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()


class Retriever(ABC):
    '''
    Retrieves features tags from a data source.
    '''

    def __init__(self):
        self.cache = {}

    @abstractmethod
    def get_sorted_timestamps_and_systems(self) -> list[dict[str, Any]]:
        '''
        Gets the complete list of all timestamps and the systems run under each timestamp.
        Format: [{"timestamp":<timestamp>,"systems":[<system1>,...,<systemN>]}]
        '''

    def get_single_json(self, timestamp: str, system: str) -> SystemFeatures:
        '''
        Given the timestamp and system, gets a single dictionary entry.

        :raises RuntimeError: when there is not exactly one entry for the system at the timestamp
        '''
        if timestamp in self.cache and 'systems' in self.cache[timestamp]:
            for s in self.cache[timestamp]['systems']:
                if s['system'] == system:
                    return s
        return self._get_single_json(timestamp, system)

    @abstractmethod
    def _get_single_json(self, timestamp: str, system: str) -> SystemFeatures:
        pass

    def get_systems(self, timestamp: str, systems: list[str] = None) -> Iterable[SystemFeatures]:
        '''
        Retrieves dictionary entries for all indicated systems at 
        the given timestamp. If systems is None, then it retrieves 
        all dictionary entries for all systems at the given timestamp.
        '''
        if timestamp in self.cache and 'systems' in self.cache[timestamp]:
            if not systems:
                return self.cache[timestamp]['systems']
            else:
                return [sys for sys in self.cache[timestamp]['systems'] if sys['system'] in systems]
            
        if systems:
            # Do not cache if only a subset of systems is specified
            # otherwise the cached list won't be complete
            return self._get_systems(timestamp, systems)
        else:
            if timestamp not in self.cache:
                self.cache[timestamp] = {}
            self.cache[timestamp]['systems'] = list(self._get_systems(timestamp, systems))
            return self.cache[timestamp]['systems']
        

    @abstractmethod
    def _get_systems(self, timestamp: str, systems: list[str] = None) -> Iterable[SystemFeatures]:
        pass

    def get_all_features(self, timestamp: str) -> dict[str, list[Any]]:
        '''
        Retrives a dictionary of all feature data at the given timestmap.
        It contains only feature keys (e.g. cmds, endpoints) and the list
        of all possible feature data at the indicated moment in time.
        '''
        if timestamp in self.cache and 'all-features' in self.cache[timestamp]:
            return self.cache[timestamp]['all-features']
        if timestamp not in self.cache:
            self.cache[timestamp] = {}
        all_feat = self._get_all_features(timestamp)
        self.cache[timestamp]['all-features'] = all_feat
        return all_feat

    @abstractmethod
    def _get_all_features(self, timestamp: str) -> dict[str, list[Any]]:
        pass

    @abstractmethod
    def close(self) -> None:
        pass


class MongoRetriever(Retriever):
    '''
    Retrieves feature data from a mongodb instance at snapd.features. 
    The mongodb entries should all be SystemFeatures documents with a 
    timestamp added. Use of this retriever, requires a credentials 
    json file with host, port, user, and password defined.
    '''

    def __init__(self, creds_file):
        super().__init__()
        config = json.load(creds_file)
        self.client = pymongo.MongoClient(
            host=config['host'], port=config['port'], username=config['user'], password=config['password'])
        self.collection = self.client.snapd.features

    def close(self):
        self.client.close()

    def get_sorted_timestamps_and_systems(self) -> list[dict[str, Any]]:
        results = self.collection.find({}, {'timestamp': 1, 'system': 1, 'all_features': 1})
        dictionary = defaultdict(list)
        for result in results:
            if 'all_features' in result and result['all_features'] == True:
                continue
            dictionary[result['timestamp'].isoformat()].append(
                result['system'])
        return [{"timestamp": entry[0], "systems": entry[1]} for entry in sorted(dictionary.items(), reverse=True)]

    def _get_systems(self, timestamp: str, systems: list[str] = None) -> Iterable[SystemFeatures]:
        if systems:
            for system in systems:
                system_jsons = self.collection.find(
                    {'timestamp': datetime.datetime.fromisoformat(timestamp), 'system': system})
                for system_json in system_jsons:
                    yield system_json
        else:
            for result in self.collection.find({'timestamp': datetime.datetime.fromisoformat(timestamp)}):
                if 'all_features' not in result or not result['all_features']:
                    yield result

    def _get_single_json(self, timestamp: str, system: str) -> SystemFeatures:
        json_result = self.collection.find(
            {'timestamp': datetime.datetime.fromisoformat(timestamp), 'system': system}).to_list()
        if len(json_result) != 1:
            raise RuntimeError(f'{len(json_result)} entries of system {system} found in collection {timestamp}')
        return json_result[0]
    
    def _get_all_features(self, timestamp) -> dict[str, list[Any]]:
        json_result = self.collection.find(
            {'timestamp': datetime.datetime.fromisoformat(timestamp), 'all_features': True}).to_list()
        if len(json_result) != 1:
            raise RuntimeError(f'{len(json_result)} entries of feature coverage information found in collection {timestamp}')
        res = json_result[0]
        del res['timestamp']
        del res['all_features']
        return res


class DirRetriever(Retriever):
    '''
    Retrieves features tagging data from the filesystem.
    It assumes data is saved in the following structure:
     <dir>/<timestamp>/<system>.json.
    To populate a directory with data from mongo, do the following:

    ./query_features.py export -f /mongo/creds.json -o /write/dir -t <timestamp1> .. <timestampN> -s <system1> .. <systemN>

    Then one can use /write/dir as a data source with this retriever
    '''

    def __init__(self, dir: str):
        super().__init__()
        if not os.path.exists(dir):
            raise RuntimeError(f'directory {dir} does not exist')
        self.dir = dir

    @staticmethod
    def __get_filename_without_last_ext(filename):
        return filename.rsplit('.', 1)[0]

    def close(self):
        pass

    def get_sorted_timestamps_and_systems(self) -> list[dict[str, Any]]:
        dictionary = defaultdict(list)
        for timestamp in os.listdir(self.dir):
            timestamp_path = os.path.join(self.dir, timestamp)
            if not os.path.isdir(timestamp_path):
                continue
            for filename in os.listdir(timestamp_path):
                if filename.endswith('.json') and not filename == ALL_FEATURES_FILE:
                    system = self.__get_filename_without_last_ext(filename)
                    dictionary[timestamp].append(system)
        return [{"timestamp": entry[0], "systems": entry[1]} for entry in sorted(dictionary.items(), reverse=True)]

    def _get_systems(self, timestamp: str, systems: list[str] = None) -> Iterable[SystemFeatures]:
        timestamp_dir = os.path.join(self.dir, timestamp)
        if not os.path.isdir(timestamp_dir):
            raise RuntimeError(
                f'timestamp {timestamp} not present in dir {self.dir}')
        for filename in os.listdir(timestamp_dir):
            if filename == ALL_FEATURES_FILE:
                continue
            if filename.endswith('.json') and (not systems or self.__get_filename_without_last_ext(filename) in systems):
                with open(os.path.join(timestamp_dir, filename), 'r', encoding='utf-8') as f:
                    yield json.load(f)

    def _get_single_json(self, timestamp: str, system: str) -> SystemFeatures:
        sys_path = os.path.join(self.dir, timestamp, system+'.json')
        if not os.path.exists(sys_path):
            raise RuntimeError(f'system file not found {sys_path}')
        with open(sys_path, 'r', encoding='utf-8') as f:
            return json.load(f)
        
    def _get_all_features(self, timestamp) -> dict[str, list[Any]]:
        sys_path = os.path.join(self.dir, timestamp, ALL_FEATURES_FILE)
        if not os.path.exists(sys_path):
            raise RuntimeError(f'{ALL_FEATURES_FILE} not found')
        with open(sys_path, 'r', encoding='utf-8') as f:
            all = json.load(f)
            if 'timestamp' in all:
                del all['timestamp']
            if 'all_features' in all:
                del all['all_features']
            return all


def consolidate_system_features(system_json: SystemFeatures, include_tasks: Iterable[TaskId] = None, exclude_tasks: Iterable[TaskId] = None) -> dict[str, list[dict[str, Any]]]:
    '''
    Given a dictionary of system feature data, consolidates features across
    all tests into one feature data dictionary.

    :param system_json: a SystemFeatures dictionary
    :param include_tasks: if not None, will only consolidate tasks present in this list
    :param exclude_tasks: if not None, will not consolidate tasks present in this list
    :returns: a dictionary with only feature data
    '''
    features = defaultdict(list)
    for test in system_json['tests']:
        if include_tasks is not None and TaskIdVariant(test['suite'], test['task_name'], test['variant']) not in include_tasks:
            continue
        if exclude_tasks is not None and TaskIdVariant(test['suite'], test['task_name'], test['variant']) in exclude_tasks:
            continue

        for feature_name in test.keys():
            if feature_name not in KNOWN_FEATURES:
                continue
            for feature in test[feature_name]:
                if feature not in features[feature_name]:
                    features[feature_name].append(feature)
    return features


def minus(first: dict[str, list], second: dict[str, list]) -> dict:
    '''
    Creates a new dictionary of first - second calculated on values.

    Ex: 
    >>> first = {'a':['b','c'],'d':['e']}
    >>> second = {'a':['c'], 'q':[]}
    >>> minus(first, second)
    {'a': ['b'], 'd': ['e']}
    '''
    minus = {}
    for feature, feature_list in first.items():
        if feature not in second:
            minus[feature] = feature_list
        else:
            m = [item for item in feature_list if item not in second[feature]]
            if m:
                minus[feature] = m
    return minus


def union(first: dict[str, list], second: dict[str, list]) -> dict[str, list]:
    '''
    Creates a new dictionary of first U second calculated on values.

    Ex: 
    >>> first = {'a':['b','c'],'d':['e']}
    >>> second = {'a':['c'], 'd':['f'], 'q':[]}
    >>> union(first, second)
    {'a': ['b', 'c], 'd': ['e','f'], 'q':[]}
    '''
    union = copy.deepcopy(first)
    for feature, feature_list in second.items():
        if feature not in union:
            union[feature] = feature_list
        else:
            for feat in feature_list:
                if feat not in union[feature]:
                    union[feature].append(feat)
    return union


def subtract_features(first: dict[str, list], second: dict[str, list], match_snap_types: bool) -> dict:
    if match_snap_types:
        return minus(first, second)
    
    diff = minus({'cmds': first['cmds'], 'ensures': first['ensures'], 'endpoints': first['endpoints']}, second)
    tasks = [af for af in first['tasks'] 
            if not any(af['kind'] == sf['kind'] and af['last_status'] == sf['last_status'] for sf in second['tasks'])]
    changes = [af for af in first['changes'] 
            if not any(af['kind'] == sf['kind'] for sf in second['changes'])]
    interfaces = [af for af in first['interfaces'] 
                if not any(af['name'] == sf['name'] for sf in second['interfaces'])]
    if tasks:
        diff['tasks'] = tasks
    if changes:
        diff['changes'] = changes
    if interfaces:
        diff['interfaces'] = interfaces
    return diff


def list_tasks(system_json: SystemFeatures, remove_failed: bool) -> set[TaskIdVariant]:
    '''
    Lists all tasks present in the SystemFeatures dictionary

    :param remove_failed: if true, will only include those tasks where success == True
    '''
    tasks = set()
    for test in system_json['tests']:
        if not remove_failed or test['success']:
            tasks.add(
                TaskIdVariant(test['suite'], test['task_name'], test['variant']))
    return tasks


def diff(retriever: Retriever, timestamp1: str, system1: str, timestamp2: str, system2: str, remove_failed: bool, only_same: bool, match_snap_types: bool = True) -> dict:
    '''
    Calculates set(system1_features) - set(system2_features), each at their
    respective timestamps. 

    :param remove_failed: if true, will remove all instances of tests where success == False
    :param only_same: if true, will only calculate difference between tests run on both systems
    '''
    system_json1 = retriever.get_single_json(timestamp1, system1)
    system_json2 = retriever.get_single_json(timestamp2, system2)

    include_tasks1 = None
    include_tasks2 = None
    if remove_failed or only_same:
        include_tasks1 = list_tasks(system_json1, remove_failed)
        include_tasks2 = list_tasks(system_json2, remove_failed)
        if only_same:
            include_tasks2 = include_tasks1 = include_tasks1.intersection(
                include_tasks2)

    features1 = consolidate_system_features(
        system_json1, include_tasks=include_tasks1)
    features2 = consolidate_system_features(
        system_json2, include_tasks=include_tasks2)
    
    return subtract_features(features1, features2, match_snap_types)


def diff_group_by_test(retriever: Retriever, timestamp1: str, system1: str, timestamp2: str, system2: str, remove_failed: bool, only_same: bool, match_snap_types: bool = True) -> dict:
    differences = diff(retriever, timestamp1, system1, timestamp2, system2, remove_failed, only_same, match_snap_types)

    test_dict = {}
    for feature, feat_list in differences.items():
        for feat in feat_list:
            results = find_feat(retriever, timestamp1, feat, remove_failed=remove_failed, system=system1, match_snap_types=match_snap_types)
            if len(results) == 0:
                raise RuntimeError(f"Something weird happened. The feature {feat} doesn't exist in {system1} at timestamp {timestamp1}")
            if len(results) != 1:
                raise RuntimeError(f'There should only be one system returned in searching for feature')
            if len(results[system1]) == 0:
                raise RuntimeError(f"A test was not found that contained feature {feat} for {system1}. This is impossible.")
            for test in results[system1]:
                test = str(TaskId(test.suite, test.task_name))
                if test not in test_dict:
                    test_dict[test] = {}
                if feature not in test_dict[test]:
                    test_dict[test][feature] = []
                if feat not in test_dict[test][feature]:
                    test_dict[test][feature].append(feat)
    return test_dict


def dict_to_string(d):
    stringed = ""
    for value in d.values():
        stringed = stringed + "-" + str(value)
    return stringed[1:]

def diff_group_by_test_csv(retriever: Retriever, timestamp1: str, system1: str, timestamp2: str, system2: str, remove_failed: bool, only_same: bool, match_snap_types: bool = True) -> str:
    differences = diff(retriever, timestamp1, system1, timestamp2, system2, remove_failed, only_same, match_snap_types)

    test_dict = {}
    for feature, feat_list in differences.items():
        for feat in feat_list:
            results = find_feat(retriever, timestamp1, feat, remove_failed=remove_failed, system=system1, match_snap_types=match_snap_types)
            if len(results) == 0:
                raise RuntimeError(f"Something weird happened. The feature {feat} doesn't exist in {system1} at timestamp {timestamp1}")
            if len(results) != 1:
                raise RuntimeError(f'There should only be one system returned in searching for feature')
            if len(results[system1]) == 0:
                raise RuntimeError(f"A test was not found that contained feature {feat} for {system1}. This is impossible.")
            for test in results[system1]:
                test = str(TaskId(test.suite, test.task_name))
                if test not in test_dict:
                    test_dict[test] = []
                test_dict[test].append(dict_to_string(feat))

    csv=""
    for test, features in test_dict.items():
        csv = f"{csv}\n{test}"
        for feature in features:
            csv = f"{csv},{feature}"
    return csv
            


def diff_all_features(retriever: Retriever, timestamp: str, system: str, remove_failed: bool) -> dict:
    '''
    Calculates set(coverage_features) - set(system_features), at the indicated timestamp. The difference
    in features for cmds, ensures, and endpoints is calculated on the complete feature. The difference
    for tasks is computed on kind and last_status. The difference for changes are computed on kind.
    The difference for interfaces is computed on name.

    :param remove_failed: if true, will remove all instances of tests where success == False
    '''
    sys_features = feat_sys(retriever, timestamp, system, remove_failed)
    all_features = retriever.get_all_features(timestamp)
    return subtract_features(all_features, sys_features, False)


def feat_sys(retriever: Retriever, timestamp: str, system: str, remove_failed: bool, suite: str = None, task: str = None, variant: str = None) -> dict:
    '''
    Calculates set(system_features), at the indicated timestamp. 

    :param remove_failed: if true, will remove all instances of tests where success == False
    '''
    system_json = retriever.get_single_json(timestamp, system)

    if suite or task or variant:
        tests = system_json['tests']
        system_json = SystemFeatures(schema_version=system_json['schema_version'] if 'schema_version' in system_json else '', 
                                     system=system_json['system'] if 'system' in system_json else '', 
                                     scenarios=system_json['scenarios'] if 'scenarios' in system_json else '',
                                     env_variables=system_json['env_variables'] if 'env_variables' in system_json else '')
        system_json['tests'] = [test for test in tests 
                                if (not suite or test['suite'] == suite) and 
                                (not task or test['task_name'] == task) and 
                                (not variant or test['variant'] == variant)]

    include_tasks = None
    if remove_failed:
        include_tasks = list_tasks(system_json, remove_failed)

    return consolidate_system_features(
        system_json, include_tasks=include_tasks)


def get_feature_name_from_feature(feat: dict) -> str:
    if 'cmd' in feat and len(feat) == 1:
        return 'cmds'
    elif 'method' in feat and 'path' in feat and (len(feat) == 2 or len(feat) == 3):
        return 'endpoints'
    elif 'manager' in feat and 'function' in feat and len(feat) == 2:
        return 'ensures'
    elif 'kind' in feat and 'last_status' in feat:
        return 'tasks'
    elif 'name' in feat:
        return 'interfaces'
    elif 'kind' in feat:
        return 'changes'
    return ''


def find_feat(retriever: Retriever, timestamp: str, feat: dict, remove_failed: bool, system: str = None, match_snap_types: bool = False) -> dict[str, TaskIdVariant]:
    '''
    Given a timestamp, a feature, and optionally a system, finds
    all tests that contain the indicated feature. If no system
    is specified, then returns all systems that contain the 
    feature combined with tests.

    A feature match happens for cmds, endpoints, and ensures
    when all of their fields match exactly.

    A feature match happens for tasks when the kind and last_status
    are identical.

    A feature match happens for changes when the kinds match.

    A feature match happens for interfaces when the names match.

    :param remove_failed: if true, will remove all instances of tests where success == False
    :param force_exact: if true, will match the entire feature when searching for tests
    :returns: dictionary where each key is a system and each value is a list of tests that contain the feature
    '''

    feat_name = get_feature_name_from_feature(feat)
    if not feat_name:
        raise RuntimeError(f'feature {feat} not a recognized feature')

    feat_in_test=lambda test: feat_name in test and feat in test[feat_name]
    if not match_snap_types:
        if feat_name == 'tasks':
            feat_in_test = lambda test: feat_name in test and any(t['kind'] == feat['kind'] and t['last_status'] == feat['last_status'] for t in test['tasks'])
        elif feat_name == 'changes':
            feat_in_test = lambda test: feat_name in test and any(c['kind'] == feat['kind'] for c in test['changes'])
        elif feat_name == 'interfaces':
            feat_in_test = lambda test: feat_name in test and any(i['name'] == feat['name'] for i in test['interfaces'])

    system_list = None
    if system:
        system_list = [system]
    system_jsons = retriever.get_systems(timestamp, systems=system_list)

    ret = {}
    for system_json in system_jsons:
        tests = [TaskIdVariant(suite=test['suite'], task_name=test['task_name'], variant=test['variant']) 
                    for test in system_json['tests'] if feat_in_test(test) and (not remove_failed or test['success'])]
        if tests:
            ret[system_json['system']] = tests
    return ret


def check_duplicate(args):
    task, system_json = args
    # Exclude all variants of the task from consolidation so that
    # it isn't flagged as a duplicate when variants of the same
    # task have identical features.
    task_id = TaskId(suite=task['suite'], task_name=task['task_name'])
    features = consolidate_system_features(
        system_json, exclude_tasks=[task_id])
    to_check = {key: value for key,
                value in task.items() if key in KNOWN_FEATURES}
    mns = minus(to_check, features)
    if to_check and not mns:
        return TaskIdVariant(suite=task['suite'], task_name=task['task_name'], variant=task['variant'])
    return None


def dup(retriever: Retriever, timestamp: str, system: str, remove_failed: bool) -> list[TaskIdVariant]:
    '''
    Returns tests whose features are completely covered by other tests in that system.

    :param remove_failed: if true, will remove all instances of tests whose success == False
    '''
    system_json = retriever.get_single_json(timestamp, system)

    if remove_failed:
        tests = system_json['tests']
        system_json = SystemFeatures(schema_version=system_json['schema_version'] if 'schema_version' in system_json else '', 
                                     system=system_json['system'] if 'system' in system_json else '', 
                                     scenarios=system_json['scenarios'] if 'scenarios' in system_json else '',
                                     env_variables=system_json['env_variables'] if 'env_variables' in system_json else '')
        system_json['tests'] = [
            task for task in tests if task['success']]

    duplicates = []
    with concurrent.futures.ProcessPoolExecutor() as executor:
        results = executor.map(
            check_duplicate, [(task, system_json) for task in system_json['tests']])
        for result in results:
            if result is not None:
                duplicates.append(result)

    return duplicates


def export(retriever: Retriever, output: str, timestamps: list[str], systems: list[str] = None):
    '''
    Writes the feature data to the output directory in format <dir>/<timestamp>/<system>.json.
    It creates one directory for each supplied timestamp, and will write one json file for each system
    in that timestamp and present in the systems list. If the systems list is None, then it will write
    all systems at each supplied timestamp.
    '''
    for timestamp in timestamps:
        os.makedirs(os.path.join(output, timestamp), exist_ok=True)
        for system_json in retriever.get_systems(timestamp, systems):
            with open(os.path.join(output, timestamp, system_json['system'] + ".json"), 'w', encoding='utf-8') as f:
                json.dump(system_json, f, cls=DateTimeEncoder)
        try:
            all_features = retriever.get_all_features(timestamp)
            with open(os.path.join(output, timestamp, ALL_FEATURES_FILE), 'w', encoding='utf-8') as f:
                json.dump(all_features, f, cls=DateTimeEncoder)
        except Exception as e:
            print(f'could not find all features at timestamp {timestamp}', file=sys.stderr)


def task_list(retriever: Retriever, timestamp: str) -> set[TaskIdVariant]:
    '''
    Given a timestamp, gets the complete set of tasks (suite, task, and variant names)
    '''
    all_data = retriever.get_systems(timestamp)
    tasks = set()
    for data in all_data:
        if 'tests' not in data:
            continue
        tasks.update(
            {TaskIdVariant(suite=d['suite'], task_name=d['task_name'], variant=d['variant']) for d in data['tests']}
        )
    return tasks


def add_data_source_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument('-f', '--file', help='json file containing creds for mongodb', type=argparse.FileType('r', encoding='utf-8'))
    parser.add_argument('-d', '--dir', help='folder containing feature data', type=str)


def add_diff_parser(subparsers: argparse._SubParsersAction) -> str:
    diff_description = '''
        Calculates feature diff between two systems: set(features_1) - set(features_2).
        You can specify either a json file with credentials for mongodb or a directory with features output.
        If using a directory, the directory format must be <dir>/<timestamp1>/<system1>.json and 
        <dir>/<timestamp2>/<system2>.json.

        By default, it will compare all features across both systems. If you wish to restrict the comparison
        to only tasks that were successful, use the --remove-failed flag. If you wish to restrict the
        comparison to only tasks that executed on both systems, use the --only-same flag.

        If you wish to create a directory from mongo data, use the export command instead of diff first.
    '''
    cmd = 'diff'
    diff: argparse.ArgumentParser = subparsers.add_parser(cmd, help='calculate diff between system features',
                                 description=diff_description, formatter_class=argparse.RawDescriptionHelpFormatter)
    add_data_source_args(diff)
    diff.add_argument('-t1', '--timestamp1', help='timestamp of first execution', type=str, required=True)
    diff.add_argument('-s1', '--system1', help='system of first execution', type=str, required=True)
    diff.add_argument('-t2', '--timestamp2', help='timestamp of second execution', type=str, required=True)
    diff.add_argument('-s2', '--system2', help='system of second execution', type=str, required=True)
    diff.add_argument('--remove-failed', help='remove all tasks that failed', action='store_true')
    diff.add_argument('--only-same', help='only compare tasks that were executed on both systems', action='store_true')
    return cmd


def add_diff_parsers(subparsers: argparse._SubParsersAction) -> tuple[str, str, str]:

    sys_description = '''
        Calculates feature diff between two systems: set(features_1) - set(features_2).
        You can specify either a json file with credentials for mongodb or a directory with features output.
        If using a directory, the directory format must be <dir>/<timestamp1>/<system1>.json and 
        <dir>/<timestamp2>/<system2>.json.

        By default, it will compare all features across both systems. If you wish to restrict the comparison
        to only tasks that were successful, use the --remove-failed flag. If you wish to restrict the
        comparison to only tasks that executed on both systems, use the --only-same flag.

        If you wish to create a directory from mongo data, use the export command instead of diff first.
    '''
    all_description = '''
        Calculates feature diff between all possible features a system's: set(all_features) - set(system_features).
        You can specify either a json file with credentials for mongodb or a directory with features output.
        If using a directory, the directory format must be <dir>/<timestamp1>/<system1>.json and 
        <dir>/<timestamp1>/all-feat.json.

        By default, it will compare all features . If you wish to restrict the comparison
        to only tasks that were successful, use the --remove-failed flag.

        If you wish to create a directory from mongo data, use the export command instead of diff first.
    '''
    cmd = 'diff'
    cmd_sys = 'systems'
    cmd_all = 'all-features'
    diff: argparse.ArgumentParser = subparsers.add_parser(cmd, help='calculate diff between features',
                                 description='Calculates feature diff either between two systems or between a system and all possible features.')
    diff_subparsers = diff.add_subparsers(dest='diff_cmd')
    sys = diff_subparsers.add_parser(cmd_sys, help='calculate diff between two systems\' features',
                               description=sys_description, formatter_class=argparse.RawDescriptionHelpFormatter)
    add_data_source_args(sys)
    sys.add_argument('-t1', '--timestamp1', help='timestamp of first execution', type=str, required=True)
    sys.add_argument('-s1', '--system1', help='system of first execution', type=str, required=True)
    sys.add_argument('-t2', '--timestamp2', help='timestamp of second execution', type=str, required=True)
    sys.add_argument('-s2', '--system2', help='system of second execution', type=str, required=True)
    sys.add_argument('--remove-failed', help='remove all tasks that failed', action='store_true')
    sys.add_argument('--only-same', help='only compare tasks that were executed on both systems', action='store_true')
    sys.add_argument('--match-snap-types', help='match the entire feature, including snap types', action='store_true')
    sys.add_argument('--sort-by-test', help='group features by the tests they are found in', action='store_true')
    sys.add_argument('--csv', help='output the diff in csv format rather than json', action='store_true')

    all = diff_subparsers.add_parser(cmd_all, help='calculate diff between system features and all features',
                                     description=all_description, formatter_class=argparse.RawDescriptionHelpFormatter)
    add_data_source_args(all)
    all.add_argument('-t', '--timestamp', help='timestamp of instance to search', required=True, type=str)
    all.add_argument('-s', '--system', help='system whose features should be searched', required=True, type=str)
    all.add_argument('--remove-failed', help='remove all tasks that failed', action='store_true')
    return cmd, cmd_sys, cmd_all


def add_dup_parser(subparsers: argparse._SubParsersAction) -> str:
    dup_description = '''
        For each task present in the indicated system under the indicated timestamp,
        calculates the difference between that task's features and the system's 
        without the task: set(task_features) - set(system_features without task).
        If the difference is ever empty, then that task is printed to console as
        a duplicate feature.

        To remove all failed tasks from consideration, add the --remove-failed flag.
    '''
    cmd = 'dup'
    duplicate: argparse.ArgumentParser = subparsers.add_parser(cmd,
                                      help='show tasks whose features are completely covered by the rest',
                                      description=dup_description,
                                      formatter_class=argparse.RawDescriptionHelpFormatter)
    add_data_source_args(duplicate)
    duplicate.add_argument('-t', '--timestamp', help='timestamp of instance to search', required=True, type=str)
    duplicate.add_argument('-s', '--system', help='system whose features should be searched', required=True, type=str)
    duplicate.add_argument('--remove-failed', help='remove all tasks that failed', action='store_true')
    return cmd


def add_export_parser(subparsers: argparse._SubParsersAction) -> str:
    cmd = 'export'
    export: argparse.ArgumentParser = subparsers.add_parser(cmd, help='export data to output local directory',
                                   description='Grabs system json files by timestamps and systems and saves them to the folder indicated in the output arguement.')
    add_data_source_args(export)
    export.add_argument('-t', '--timestamps', help='space-separated list of identifying timestamps', required=True, nargs='+')
    export.add_argument('-s', '--systems', help='space-separated list of systems', nargs='*')
    export.add_argument('-o', '--output', help='folder to save feature data', required=True, type=str)
    return cmd


def add_list_parser(subparsers: argparse._SubParsersAction) -> str:
    cmd = 'list'
    lst: argparse.ArgumentParser = subparsers.add_parser(cmd, help='lists all timestamps with systems present in data source',
                                description='Lists all timestamps with systems present in data source.')
    add_data_source_args(lst)
    return cmd


def add_all_features_parser(subparsers: argparse._SubParsersAction) -> tuple[str, str, str, str]:
    cmd = 'feat'
    cmd_all = 'all'
    cmd_sys = 'sys'
    cmd_find = 'find'
    feat: argparse.ArgumentParser = subparsers.add_parser(cmd, help='', description='')
    feat_subparsers = feat.add_subparsers(dest='features_cmd')
    all = feat_subparsers.add_parser(cmd_all, help='lists all features at a given timestamp',
                                     description='Lists all features for a given timestamp present in data source.')
    add_data_source_args(all)
    all.add_argument('-t', '--timestamp', help='timestamp for feature data', required=True, type=str)

    sys = feat_subparsers.add_parser(cmd_sys, help='lists all features for a given system and timestamp',
                                     description='Lists all features for a given system and timestamp present in data source.')
    add_data_source_args(sys)
    sys.add_argument('-t', '--timestamp', help='timestamp for feature data', required=True, type=str)
    sys.add_argument('-s', '--system', help='system for feature data', required=True, type=str)
    sys.add_argument('--suite', help='if provided, only grab features from this suite', default=None, type=str)
    sys.add_argument('--task', help='if provided, only grab features of this task', default=None, type=str)
    sys.add_argument('--variant', help='if provided, only grab features with this variant', default=None, type=str)
    sys.add_argument('--remove-failed', help='remove all tasks that failed', action='store_true')


    find = feat_subparsers.add_parser(cmd_find, help='given a feature, finds which tests contain it',
                                     description='Lists tests that contain the feature')
    add_data_source_args(find)
    find.add_argument('-t', '--timestamp', help='timestamp for feature data', required=True, type=str)
    find.add_argument('-s', '--system', help='(optional) system to search for feature in', default=None, type=str)
    find.add_argument('--feat', help='feature to search for (json format)', required=True, type=str)
    find.add_argument('--remove-failed', help='remove all tasks that failed', action='store_true')
    find.add_argument('--match-snap-types', help='match the entire feature, including snap types', action='store_true')
    return cmd, cmd_all, cmd_sys, cmd_find


def main():
    parser = argparse.ArgumentParser(
        description='cli to query data source containing feature data')
    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True
    diff_cmd, diff_sys_cmd, diff_all_cmd = add_diff_parsers(subparsers)
    dup_cmd = add_dup_parser(subparsers)
    export_cmd = add_export_parser(subparsers)
    list_cmd = add_list_parser(subparsers)
    feat_cmd, feat_all_cmd, feat_sys_cmd, feat_find_cmd = add_all_features_parser(subparsers)

    args = parser.parse_args()

    retriever_creator = None
    if args.dir:
        def retriever_creator(): return DirRetriever(args.dir)
    elif args.file:
        def retriever_creator(): return MongoRetriever(args.file)
    else:
        raise RuntimeError(
            'you must specify either a mongodb credential file (-f) or a directory with feature tagging results (-d)')

    with closing(retriever_creator()) as retriever:
        if args.command == diff_cmd:
            if args.diff_cmd == diff_sys_cmd:
                if args.sort_by_test:
                    if args.csv:
                        result = diff_group_by_test_csv(retriever, args.timestamp1, args.system1,
                            args.timestamp2, args.system2, args.remove_failed, args.only_same, args.match_snap_types)    
                        print(result)
                        exit(0)
                    result = diff_group_by_test(retriever, args.timestamp1, args.system1,
                            args.timestamp2, args.system2, args.remove_failed, args.only_same, args.match_snap_types)
                else:
                    result = diff(retriever, args.timestamp1, args.system1,
                            args.timestamp2, args.system2, args.remove_failed, args.only_same, args.match_snap_types)
            elif args.diff_cmd == diff_all_cmd:
                result = diff_all_features(retriever, args.timestamp, args.system, args.remove_failed)
            else:
                raise RuntimeError(f'diff command not recognized: {args.diff_cmd}')
            json.dump(result, sys.stdout, cls=DateTimeEncoder)
            print()
        elif args.command == dup_cmd:
            results = dup(retriever, args.timestamp,
                          args.system, args.remove_failed)
            if results:
                json.dump(results, sys.stdout, default=lambda x: str(x))
                print()
        elif args.command == export_cmd:
            export(retriever, args.output, args.timestamps, args.systems)
        elif args.command == list_cmd:
            result = retriever.get_sorted_timestamps_and_systems()
            json.dump(result, sys.stdout, cls=DateTimeEncoder)
            print()
        elif args.command == feat_cmd:
            if args.features_cmd == feat_all_cmd:
                result = retriever.get_all_features(args.timestamp)
                json.dump(result, sys.stdout, cls=DateTimeEncoder)
            elif args.features_cmd == feat_sys_cmd:
                result = feat_sys(retriever, args.timestamp, args.system, args.remove_failed, args.suite, args.task, args.variant)
                json.dump(result, sys.stdout, cls=DateTimeEncoder)
            elif args.features_cmd == feat_find_cmd:
                try:
                    feat = json.loads(args.feat)
                    result = find_feat(retriever, args.timestamp, feat, args.remove_failed, args.system, args.match_snap_types)
                    json.dump(result, sys.stdout, default=lambda x: str(x))
                except Exception as e:
                    raise RuntimeError(f'Error parsing feature {args.feat}: {e}')
            else:
                raise RuntimeError(f'unrecognized feature command {args.features_cmd}')
            print()
        else:
            raise RuntimeError(f'command not recognized: {args.command}')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

from features import Change, Cmd, Endpoint, Interface, Status, Task
import argparse
import json
import os
import sys

sys.path.append(os.path.dirname(os.path.abspath(__file__)))


def translate_all_features(all_features_json: str, output_json: str):
    with open(all_features_json, "r", encoding="utf-8") as f:
        all_features = json.load(f)

    if any(
        key not in all_features
        for key in [
            "endpoints",
            "tasks",
            "commands",
            "ensures",
            "interfaces",
            "changes",
        ]
    ):
        raise RuntimeError(f"missing key in all features file {all_features_json}")

    all_endpoints = []
    for endpoint in all_features["endpoints"]:
        if "actions" in endpoint and endpoint["actions"]:
            for action in endpoint["actions"]:
                all_endpoints.append(
                    Endpoint(
                        method=endpoint["method"], path=endpoint["path"], action=action
                    )
                )
        else:
            all_endpoints.append(
                Endpoint(method=endpoint["method"], path=endpoint["path"])
            )
    all_cmds = [Cmd(cmd=cmd) for cmd in all_features["commands"]]
    all_tasks = []
    for task in all_features["tasks"]:
        status_list = (
            [Status.done, Status.undone, Status.error]
            if "has-undo" in task and task["has-undo"]
            else [Status.done, Status.error]
        )
        for status in status_list:
            all_tasks.append(Task(kind=task["kind"], last_status=status))
    all_changes = [Change(kind=change) for change in all_features["changes"]]
    all_interfaces = [Interface(name=iface) for iface in all_features["interfaces"]]

    with open(output_json, "w", encoding="utf-8") as f:
        json.dump(
            {
                "cmds": all_cmds,
                "ensures": all_features["ensures"],
                "tasks": all_tasks,
                "changes": all_changes,
                "endpoints": all_endpoints,
                "interfaces": all_interfaces,
            },
            f,
        )


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Translates feature tagging output from 'snap debug features' to the same format as feature tagging reports"
    )
    parser.add_argument(
        "-f", "--file", help="snap debug features output file", required=True, type=str
    )
    parser.add_argument(
        "-o",
        "--output",
        help="where to write translated output",
        required=True,
        type=str,
    )
    args = parser.parse_args()
    translate_all_features(args.file, args.output)



#!/usr/bin/env python3

from gi.repository import GLib
import dbus
import dbus.service

from dbus.mainloop.glib import DBusGMainLoop

INTERFACE = "com.ubuntu.location.Service"
PATH = "/com/ubuntu/location/Service"

DBusGMainLoop(set_as_default=True)


class DBusProvider(dbus.service.Object):
    def __init__(self):
        bus = dbus.SystemBus()
        bus_name = dbus.service.BusName(INTERFACE, bus=bus)
        dbus.service.Object.__init__(self, bus_name, PATH)

    @dbus.service.method(dbus_interface=INTERFACE, out_signature="s")
    def AddProvider(self):
        return "location-provider-added"


if __name__ == "__main__":
    DBusProvider()
    loop = GLib.MainLoop()
    loop.run()



#!/usr/bin/env python3

import argparse
import os
import struct

parser = argparse.ArgumentParser(description='show the title of the first UEFI boot entry')
parser.add_argument("--current", action='store_true', help='show the current boot entry instead')
parser.add_argument("--path", action='store_true', help='show path instead of title')
parser.add_argument("vardir", nargs='?', default='/sys/firmware/efi/efivars',
                    help='the path to the variable directory (default: /sys/firmware/efi/efivars)')
args = parser.parse_args()

u32_f = '@I'
u16_f = '@H'
device_path_header_f = '@BBH'

def file_read(fd, fmt):
    buf = fd.read(struct.calcsize(fmt))
    if len(buf) == 0:
        return None
    return struct.unpack(fmt, buf)

order = []
with open(os.path.join(args.vardir, 'BootOrder-8be4df61-93ca-11d2-aa0d-00e098032b8c'), 'br') as b:
    file_read(b, u32_f)
    while True:
        r = file_read(b, u16_f)
        if r is None:
            break
        order.append(r[0])

with open(os.path.join(args.vardir, 'BootCurrent-8be4df61-93ca-11d2-aa0d-00e098032b8c'), 'br') as b:
    file_read(b, u32_f)
    r = file_read(b, u16_f)
    current = r[0]

if args.current:
    selected = current
else:
    selected = order[0]

with open(os.path.join(args.vardir, 'Boot{:04x}-8be4df61-93ca-11d2-aa0d-00e098032b8c'.format(selected)), 'br') as e:
    file_read(e, u32_f)
    file_read(e, u32_f)
    path_len = file_read(e, u16_f)[0]
    title_words = []
    while True:
        b = file_read(e, u16_f)
        if b[0] == 0:
            break
        title_words.append(struct.pack(u16_f, b[0]))
    last_path = None
    path_read = 0
    while path_read < path_len:
        dp_header = file_read(e, device_path_header_f)
        dp_data = e.read(dp_header[2] - struct.calcsize(device_path_header_f))
        path_read += dp_header[2]
        if dp_header[0] == 4 and dp_header[1] == 4:
            last_path = dp_data

if args.path:
    print(last_path.decode('utf-16').rstrip('\0'))
else:
    print(b''.join(title_words).decode('utf-16'))




import argparse
from copy import deepcopy
from datetime import datetime
from io import StringIO
import json
import os
from pathlib import Path
import pytest
import sys
import tempfile
from typing import Iterable
import unittest
from unittest.mock import Mock, patch
# To ensure the unit test can be run from any point in the filesystem,
# add parent folder to path to permit relative imports
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

import query_features
from query_features import TaskIdVariant
from features import SystemFeatures, TaskFeatures, Cmd, Endpoint, Change, Task, Ensure, Interface


class DictRetriever(query_features.Retriever):
    def __init__(self, data, all_features=None):
        super().__init__()
        self.data = data
        self.all_features = all_features

    def close(self):
        pass

    def get_sorted_timestamps_and_systems(self) -> list[dict[str, list[str]]]:
        pass

    def _get_single_json(self, timestamp: str, system: str) -> SystemFeatures:
        return deepcopy(self.data[timestamp][system])

    def _get_systems(self, timestamp: str, systems: list[str] = None) -> Iterable[SystemFeatures]:    
        if systems:
            return [deepcopy(self.data[timestamp][system]) for system in systems]
        else:
            return [deepcopy(data) for _, data in self.data[timestamp].items()]

    def _get_all_features(self, timestamp):
        return deepcopy(self.all_features[timestamp])


class FakeClient:
    def close(self):
        pass


class FakeCollectionReturn:
    def __init__(self, l):
        self.l = l

    def __iter__(self):
        return iter(self.l)

    def __len__(self):
        return len(self.l)

    def to_list(self):
        return self.l


class FakeMongoCollection:
    def __init__(self, list_json):
        self.list_json = list_json

    def find(self, dictionary=None, projection=None):
        l = []
        for doc in self.list_json:
            if not dictionary or all(FakeMongoCollection.check_equals(key, dictionary, doc) for key in dictionary.keys()):
                l.append(doc)
        return FakeCollectionReturn(l)
    
    def check_equals(key, dict1, dict2):
        if key not in dict1 or key not in dict2:
            return False
        ts1 = dict1[key]
        ts2 = dict2[key]
        if key != 'timestamp':
            return ts1 == ts2
        if isinstance(ts1, str):
            ts1 = datetime.fromisoformat(ts1)
        if isinstance(ts2, str):
            ts2 = datetime.fromisoformat(ts2)
        return ts1 == ts2
            


class MongoMocker:
    def __init__(self, collection_data, do_patch_stdout=False):
        self.collection_data = collection_data
        for data in self.collection_data:
            if 'timestamp' in data:
                data['timestamp'] = datetime.fromisoformat(data['timestamp'])
        self.do_patch_stdout = do_patch_stdout
        self.patch_stdout = None
        self.patch_mongo = None

    def get_stdout(self):
        return self.stdout.getvalue()

    def __enter__(self):
        data = self.collection_data

        def my_init(self, *_):
            super(query_features.MongoRetriever, self).__init__()
            self.collection = FakeMongoCollection(data)
            self.client = FakeClient()
        self.patch_mongo = patch.object(
            query_features.MongoRetriever, '__init__', my_init)
        self.patch_mongo.start()
        if self.do_patch_stdout:
            self.patch_stdout = patch('sys.stdout', new=StringIO())
            self.stdout = self.patch_stdout.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.patch_mongo:
            self.patch_mongo.stop()
        if self.patch_stdout:
            self.patch_stdout.stop()


class DirMocker:
    def __init__(self, collection_data, do_patch_stdout=False):
        self.collection_data = collection_data
        self.do_patch_stdout = do_patch_stdout
        self.patch_stdout = None
        self.tmpdir = None

    def get_stdout(self):
        return self.stdout.getvalue()

    def get_dir(self):
        return self.tmpdir.name

    def __populate_dir(self):
        for doc in self.collection_data:
            dir = os.path.join(self.tmpdir.name, doc['timestamp'])
            os.makedirs(dir, exist_ok=True)
            if 'all_features' in doc:
                with open(os.path.join(dir, "all-features.json"), 'w', encoding='utf-8') as f:
                    json.dump(doc, f)
            else:
                with open(os.path.join(dir, f'{doc["system"]}.json'), 'w', encoding='utf-8') as f:
                    json.dump(doc, f)

    def __enter__(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.__populate_dir()
        if self.do_patch_stdout:
            self.patch_stdout = patch('sys.stdout', new=StringIO())
            self.stdout = self.patch_stdout.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.tmpdir.cleanup()
        if self.patch_stdout:
            self.patch_stdout.stop()


class TestQueryFeatures:

    def test_dirretriever_get_sorted_timestamps_and_systems(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.mkdir(os.path.join(tmpdir, 'timestamp1'))
            Path(os.path.join(tmpdir, 'randomFile')).touch()
            os.mkdir(os.path.join(tmpdir, 'timestamp2'))
            Path(os.path.join(tmpdir, 'timestamp1', 'system1.json')).touch()
            Path(os.path.join(tmpdir, 'timestamp1', 'system2.json')).touch()
            Path(os.path.join(tmpdir, 'timestamp2', 'system2.json')).touch()
            Path(os.path.join(tmpdir, 'timestamp2', 'randomfile')).touch()
            os.mkdir(os.path.join(tmpdir, 'timestamp3'))

            retriever = query_features.DirRetriever(tmpdir)
            results = retriever.get_sorted_timestamps_and_systems()
            assert 2 == len(results)
            assert {'timestamp':'timestamp2','systems':['system2']} in results
            assert {'timestamp':'timestamp1','systems':['system1','system2']} in results or {'timestamp':'timestamp1','systems':['system2','system1']} in results


    def test_dirretriever_get_systems(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.mkdir(os.path.join(tmpdir, 'timestamp1'))
            with open(os.path.join(tmpdir, 'timestamp1', 'system1.json'), 'w', encoding='utf-8') as f:
                json.dump(SystemFeatures(system='system1'), f)
            with open(os.path.join(tmpdir, 'timestamp1', 'system2.json'), 'w', encoding='utf-8') as f:
                json.dump(SystemFeatures(system='system2'), f)

            retriever = query_features.DirRetriever(tmpdir)
            results = list(retriever.get_systems('timestamp1'))
            assert 2 == len(results)
            assert SystemFeatures(system='system1') in results
            assert SystemFeatures(system='system2') in results
            results = list(retriever.get_systems('timestamp1', ['system1', 'system2']))
            assert 2 == len(results)
            assert SystemFeatures(system='system1') in results
            assert SystemFeatures(system='system2') in results
            results = list(retriever.get_systems('timestamp1', ['system1']))
            assert [SystemFeatures(system='system1')] == results

    def test_dirretriever_get_single_json(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.mkdir(os.path.join(tmpdir, 'timestamp1'))
            with open(os.path.join(tmpdir, 'timestamp1', 'system1.json'), 'w', encoding='utf-8') as f:
                json.dump(SystemFeatures(system='system1'), f)
            retriever = query_features.DirRetriever(tmpdir)
            result = retriever.get_single_json('timestamp1', 'system1')
            assert SystemFeatures(system='system1') == result

    def test_dirretriever_get_all_features(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            os.mkdir(os.path.join(tmpdir, 'timestamp1'))
            all_features = {
                'timestamp':'timestamp1',
                'cmds':[Cmd(cmd='snap list'),Cmd(cmd='snap pack')],
                'changes':[Change(kind='refresh',snap_types=[])],
                'tasks':[Task(kind='link',snap_types=['snapd'],last_status='Done')]}
            with open(os.path.join(tmpdir, 'timestamp1', 'all-features.json'), 'w', encoding='utf-8') as f:
                json.dump(all_features, f)
            retriever = query_features.DirRetriever(tmpdir)
            result = retriever.get_all_features('timestamp1')
            del all_features['timestamp']
            assert all_features == result

    def test_consolidate_features(self):
        j = {"tests": [
            {"task_name": "task1",
             "cmds": [{"cmd": "snap list --all"}, {"cmd": "snap ack file"},],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc1"}]},
            {"task_name": "task2",
             "cmds": [{"cmd": "snap do things"}, {"cmd": "snap list --all"}],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc2"}]
             }
        ]}
        c = query_features.consolidate_system_features(j)
        assert len(c) == 2
        assert "cmds" in c
        assert len(c["cmds"]) == 3
        assert {"cmd": "snap list --all"} in c["cmds"]
        assert {"cmd": "snap ack file"} in c["cmds"]
        assert {"cmd": "snap do things"} in c["cmds"]
        assert "ensures" in c
        assert len(c["ensures"]) == 2
        assert {"manager":"SnapManager","function":"ensureFunc1"} in c["ensures"]
        assert {"manager":"SnapManager","function":"ensureFunc2"} in c["ensures"]


    def test_consolidate_features_exclude_task(self):
        j = {"tests": [
            {"suite": "suite", "task_name": "task1", "variant": "a",
             "cmds": [{"cmd": "snap list --all"}, {"cmd": "snap ack file"},],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc1"}]},
            {"suite": "suite", "task_name": "task2", "variant": "",
             "cmds": [{"cmd": "snap do things"}, {"cmd": "snap list --all"}],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc2"}]
             }
        ]}
        c = query_features.consolidate_system_features(j, exclude_tasks=[query_features.TaskId(suite='suite',task_name="task1")])
        assert len(c) == 2
        assert "cmds" in c
        assert len(c["cmds"]) == 2
        assert {"cmd": "snap list --all"} in c["cmds"]
        assert {"cmd": "snap do things"} in c["cmds"]
        assert "ensures" in c
        assert len(c["ensures"]) == 1
        assert {"manager": "SnapManager", "function": "ensureFunc2"} in c["ensures"]

    def test_consolidate_features_include_task(self):
        j = {"tests": [
            {"suite": "suite", "task_name": "task1", "variant": "a",
             "cmds": [{"cmd": "snap list --all"}, {"cmd": "snap ack file"},],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc1"}]},
            {"suite": "suite", "task_name": "task2", "variant": "",
             "cmds": [{"cmd": "snap do things"}, {"cmd": "snap list --all"}],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc2"}]
             }
        ]}
        c = query_features.consolidate_system_features(j, include_tasks=[query_features.TaskId(suite='suite',task_name="task2")])
        assert len(c) == 2
        assert "cmds" in c
        assert len(c["cmds"]) == 2
        assert {"cmd": "snap list --all"} in c["cmds"]
        assert {"cmd": "snap do things"} in c["cmds"]
        assert "ensures" in c
        assert len(c["ensures"]) == 1
        assert {"manager": "SnapManager", "function": "ensureFunc2"} in c["ensures"]

    def test_features_minus(self):
        j = {"cmds": [{"cmd": "snap list --all"}, {"cmd": "snap ack file"},],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc1"}],
             }
        k = {"cmds": [{"cmd": "snap list --all"}],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc2"}],
             }
        minus = query_features.minus(j, k)
        assert len(minus) == 2
        assert "cmds" in minus
        assert "ensures" in minus
        assert len(minus["cmds"]), 1
        assert {"cmd": "snap ack file"} in minus["cmds"]
        assert len(minus["ensures"]) == 1
        assert {"manager": "SnapManager", "function": "ensureFunc1"} in minus["ensures"]


    def test_subract_features_no_match_snap_types(self):
        j = {"cmds": [{"cmd": "snap list --all"}, {"cmd": "snap ack file"},],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc1"}],
             "interfaces": [Interface(name="my-interf", plug_snap_type="aaa", slot_snap_type="bbb")],
             "endpoints": [Endpoint(method="GET", path="/my-end", action="")],
             "tasks": [Task(kind="my-task", last_status="Done", snap_types=["aaa"]), 
                       Task(kind="my-task", last_status="Error", snap_types=["bbb"])],
             "changes": [Change(kind="my-task", snap_types=["aaa"])]
             }
        k = {"cmds": [{"cmd": "snap list --all"}],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc2"}],
             "interfaces": [Interface(name="my-interf", plug_snap_type="bbb", slot_snap_type="aaa")],
             "endpoints": [Endpoint(method="POST", path="/my-end", action="")],
             "tasks": [Task(kind="my-task", last_status="Done", snap_types=["bbb"]), 
                       Task(kind="my-task", last_status="Error", snap_types=["bbb"])],
             "changes": [Change(kind="my-task", snap_types=["bbb"])]
             }
        minus = query_features.subtract_features(j, k, False)
        assert len(minus) == 3
        assert "cmds" in minus
        assert "ensures" in minus
        assert "endpoints" in minus
        assert len(minus["cmds"]), 1
        assert {"cmd": "snap ack file"} in minus["cmds"]
        assert len(minus["ensures"]) == 1
        assert {"manager": "SnapManager", "function": "ensureFunc1"} in minus["ensures"]
        assert len(minus["endpoints"]), 1
        assert Endpoint(method="GET", path="/my-end", action="") in minus['endpoints']
    

    def test_subtract_features_match_snap_types(self):
        j = {"cmds": [{"cmd": "snap list --all"}, {"cmd": "snap ack file"},],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc1"}],
             "interfaces": [Interface(name="my-interf", plug_snap_type="aaa", slot_snap_type="bbb")],
             "endpoints": [Endpoint(method="GET", path="/my-end", action="")],
             "tasks": [Task(kind="my-task", last_status="Done", snap_types=["aaa"]), 
                       Task(kind="my-task", last_status="Error", snap_types=["bbb"])],
             "changes": [Change(kind="my-task", snap_types=["aaa"])]
             }
        k = {"cmds": [{"cmd": "snap list --all"}],
             "ensures": [{"manager": "SnapManager", "function": "ensureFunc2"}],
             "interfaces": [Interface(name="my-interf", plug_snap_type="bbb", slot_snap_type="aaa")],
             "endpoints": [Endpoint(method="POST", path="/my-end", action="")],
             "tasks": [Task(kind="my-task", last_status="Done", snap_types=["bbb"]), 
                       Task(kind="my-task", last_status="Error", snap_types=["bbb"])],
             "changes": [Change(kind="my-task", snap_types=["bbb"])]
             }
        minus = query_features.subtract_features(j, k, True)
        assert len(minus) == 6
        assert "cmds" in minus
        assert "ensures" in minus
        assert "endpoints" in minus
        assert "tasks" in minus
        assert "changes" in minus
        assert "interfaces" in minus
        assert len(minus["cmds"]), 1
        assert {"cmd": "snap ack file"} in minus["cmds"]
        assert len(minus["ensures"]) == 1
        assert {"manager": "SnapManager", "function": "ensureFunc1"} in minus["ensures"]
        assert len(minus["endpoints"]), 1
        assert Endpoint(method="GET", path="/my-end", action="") in minus['endpoints']
        assert len(minus["interfaces"]), 1
        assert j['interfaces'][0] in minus['interfaces']
        assert len(minus["tasks"]), 1
        assert j['tasks'][0] in minus['tasks']
        assert len(minus["changes"]), 1
        assert j['changes'][0] in minus['changes']


    def test_list_tasks(self):
        sys_json = SystemFeatures(tests=[
            TaskFeatures(success=True,task_name='task1',variant='variant1',suite='suite1'),
            TaskFeatures(success=False,task_name='task2',variant='variant2',suite='suite2'),
        ])
        tasks_all = query_features.list_tasks(sys_json, False)
        tasks_success = query_features.list_tasks(sys_json, True)
        assert {query_features.TaskIdVariant('suite1', 'task1', 'variant1'),
                             query_features.TaskIdVariant('suite2', 'task2', 'variant2')} == tasks_all
        assert {query_features.TaskIdVariant('suite1', 'task1', 'variant1')} == tasks_success

    def test_list_tasks_empty(self):
        sys_json = SystemFeatures(tests=[])
        tasks = query_features.list_tasks(sys_json, False)
        assert set() == tasks

    def test_check_dup_none(self):
        system_json = SystemFeatures(tests=[
            TaskFeatures(suite='suite',task_name='task1',variant='a',cmds=[Cmd(cmd='cmd1'),Cmd(cmd='cmd2')]),
            TaskFeatures(suite='suite',task_name='task1',variant='b',cmds=[Cmd(cmd='cmd1'),Cmd(cmd='cmd2')]),
            TaskFeatures(suite='suite',task_name='task2',variant='',endpoints=[Endpoint(method='GET', path='/v2/snaps')]),
            TaskFeatures(suite='suite',task_name='task3',variant='',endpoints=[Endpoint(method='GET', path='/v2/snaps')]),
            TaskFeatures(suite='suite',task_name='task4',variant='b',cmds=[Cmd(cmd='cmd1')]),
        ])
        dups = query_features.check_duplicate((system_json['tests'][0], system_json))
        assert dups == None
        dups = query_features.check_duplicate((system_json['tests'][1], system_json))
        assert dups == None

    def test_check_dup_no_variant(self):
        system_json = SystemFeatures(tests=[
            TaskFeatures(suite='suite',task_name='task1',variant='a',cmds=[Cmd(cmd='cmd1'),Cmd(cmd='cmd2')]),
            TaskFeatures(suite='suite',task_name='task1',variant='b',cmds=[Cmd(cmd='cmd1'),Cmd(cmd='cmd2')]),
            TaskFeatures(suite='suite',task_name='task2',variant='',endpoints=[Endpoint(method='GET', path='/v2/snaps')]),
            TaskFeatures(suite='suite',task_name='task3',variant='',endpoints=[Endpoint(method='GET', path='/v2/snaps')]),
            TaskFeatures(suite='suite',task_name='task4',variant='b',cmds=[Cmd(cmd='cmd1')]),
        ])
        dups = query_features.check_duplicate((system_json['tests'][2], system_json))
        assert query_features.TaskIdVariant(suite='suite',task_name='task2',variant='') == dups
        dups = query_features.check_duplicate((system_json['tests'][3], system_json))
        assert query_features.TaskIdVariant(suite='suite',task_name='task3',variant='') == dups
        dups = query_features.check_duplicate((system_json['tests'][4], system_json))
        assert query_features.TaskIdVariant(suite='suite',task_name='task4',variant='b') == dups


    def test_dup(self):
        data = {'timestamp1': {'system1': {'tests': [
            TaskFeatures(suite='suite', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list --all"),
                               Cmd(cmd="snap ack file")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack file"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]),
            TaskFeatures(suite='suite', task_name='task3', success=False, variant='v2',
                         cmds=[
                             Cmd(cmd="snap pack file")],
                         endpoints=[Endpoint(method="GET", path="/v2/snaps")]),]}}}
        retriever = DictRetriever(data)
        dup = query_features.dup(retriever, 'timestamp1', 'system1', False)
        assert [query_features.TaskIdVariant(suite='suite', task_name='task3', variant='v2')] == dup
        dup = query_features.dup(retriever, 'timestamp1', 'system1', True)
        assert [] == dup

    def test_dup_variants(self):
        data = {'timestamp1': {'system1': {'tests': [
            TaskFeatures(suite='suite', task_name='task1', success=True, variant='a',
                         cmds=[Cmd(cmd="snap list --all"),
                               Cmd(cmd="snap ack file")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite', task_name='task1', success=True, variant='b',
                         cmds=[Cmd(cmd="snap list --all"),
                               Cmd(cmd="snap ack file")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite', task_name='task3', success=False, variant='v2',
                         endpoints=[Endpoint(method="GET", path="/v2/snaps")]),]}}}
        retriever = DictRetriever(data)
        dup = query_features.dup(retriever, 'timestamp1', 'system1', False)
        # Features from variants of the same test should not influence
        # duplicate calculation. The only duplicate task should be task3
        assert [query_features.TaskIdVariant(
            suite='suite', task_name='task3', variant='v2')] == dup
        dup = query_features.dup(retriever, 'timestamp1', 'system1', True)
        assert [] == dup

    def test_export(self):
        t1s1_dict = SystemFeatures(system='system1', tests=[
            TaskFeatures(suite='suite', task_name='task1',
                         success=True, variant=''),
            TaskFeatures(suite='suite', task_name='task2', success=True, variant='v1')])
        t2s1_dict = SystemFeatures(system='system1', tests=[
            TaskFeatures(suite='suite', task_name='task1', success=False, variant='')])
        s2_dict = SystemFeatures(system='system2')
        data = {'timestamp1': {'system1': t1s1_dict,
                               'system2': s2_dict},
                'timestamp2': {'system1': t2s1_dict,
                               'system2': s2_dict, }}
        retriever = DictRetriever(data)

        def check_equal(file, ref_dict):
            assert os.path.isfile(file)
            with open(file, 'r', encoding='utf-8') as f:
                assert ref_dict == json.load(f)

        with tempfile.TemporaryDirectory() as tmpdir:
            query_features.export(retriever, tmpdir, ['timestamp1', 'timestamp2'], None)
            timestamp1 = os.path.join(tmpdir, 'timestamp1')
            timestamp2 = os.path.join(tmpdir, 'timestamp2')
            assert os.path.isdir(timestamp1)
            assert os.path.isdir(timestamp2)
            check_equal(os.path.join(timestamp1, 'system1.json'), t1s1_dict)
            check_equal(os.path.join(timestamp1, 'system2.json'), s2_dict)
            check_equal(os.path.join(timestamp2, 'system1.json'), t2s1_dict)
            check_equal(os.path.join(timestamp2, 'system2.json'), s2_dict)

    def test_diff(self):
        data = {'timestamp1': {'system1': {'tests': [
            TaskFeatures(suite='suite', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list --all"),
                               Cmd(cmd="snap ack file")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack file"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")])]},
            'system2': {'tests': []}},
            'timestamp2': {'system1': {'tests': [
                TaskFeatures(suite='suite', task_name='task1', success=False, variant='',
                             cmds=[
                                 Cmd(cmd="snap list --all")],
                             endpoints=[Endpoint(
                                 method="GET", path="/v2/changes/{id}"), Endpoint(method="GET", path="/v2/snaps")],
                             changes=[Change(kind="install-snap", snap_types=["app"])])]},
                           'system2': {'tests': []}}}
        retriever = DictRetriever(data)

        # When getting difference only between the same tasks in both systems,
        # the only difference is in suite:task1
        diff = query_features.diff(retriever, 'timestamp1', 'system1', 'timestamp2', 'system1', False, True)
        assert {"cmds":[Cmd(cmd="snap ack file")]} == diff
        diff = query_features.diff(retriever, 'timestamp2', 'system1', 'timestamp1', 'system1', False, True)
        assert {"endpoints":[Endpoint(method="GET", path="/v2/changes/{id}")]} == diff

        # When getting difference only between the same tasks in both systems,
        # and also removing failed tasks, then there are no tasks features to compare
        # since suite:task1 failed on timestamp2
        diff = query_features.diff(retriever, 'timestamp1', 'system1', 'timestamp2', 'system1', True, True)
        assert {} == diff
        diff = query_features.diff(retriever, 'timestamp2', 'system1', 'timestamp1', 'system1', True, True)
        assert {} == diff

        # When getting all differences, suite:task2:v1, that isn't present
        # on the timestamp2 run, gets counted.
        diff = query_features.diff(retriever, 'timestamp1', 'system1', 'timestamp2', 'system1', False, False)
        assert {"cmds":[Cmd(cmd="snap ack file"),Cmd(cmd="snap pack file"),Cmd(cmd="snap debug api")],
                              "endpoints":[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]} == diff
        diff = query_features.diff(retriever, 'timestamp2', 'system1', 'timestamp1', 'system1', False, False)
        assert {"endpoints":[Endpoint(method="GET", path="/v2/changes/{id}")]} == diff

        # When removing all failed executions, the difference becomes
        # all features of suite:task1 and suite:task2:v1 in timestamp1
        # because suite:task1 failed in timestamp2 and so is removed.
        diff = query_features.diff(retriever, 'timestamp1', 'system1', 'timestamp2', 'system1', True, False)
        assert {
            "cmds":[Cmd(cmd="snap list --all"),Cmd(cmd="snap ack file"),Cmd(cmd="snap pack file"),Cmd(cmd="snap debug api")],
            "endpoints":[Endpoint(method="GET", path="/v2/snaps"),Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")],
            "changes":[Change(kind="install-snap", snap_types=["app"])]
        } == diff
        diff = query_features.diff(retriever, 'timestamp1', 'system1', 'timestamp1', 'system1', True, False)
        assert {} == diff

        # Empty dictionaries of tests have no feature difference
        diff = query_features.diff(retriever, 'timestamp1', 'system2', 'timestamp2', 'system2', False, False)
        assert {} == diff
        diff = query_features.diff(retriever, 'timestamp2', 'system2', 'timestamp1', 'system2', False, False)
        assert {} == diff

    def test_feat_sys(self):
        data = {'timestamp1': {'system': {'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap routine file-access")],
                         endpoints=[Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]),
                         ]}}}
        retriever = DictRetriever(data)

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', False)
        expected = {'cmds':[cmd for entry in data['timestamp1']['system']['tests'] if 'cmds' in entry for cmd in entry['cmds']],
                    'endpoints':[endpt for entry in data['timestamp1']['system']['tests'] if 'endpoints' in entry for endpt in entry['endpoints']],
                    'changes':[change for entry in data['timestamp1']['system']['tests'] if 'changes' in entry for change in entry['changes']]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', True)
        expected = {'cmds':[Cmd(cmd="snap list"),Cmd(cmd="snap pack"),Cmd(cmd="snap debug api")],
                    'endpoints':[Endpoint(method="GET", path="/v2/snaps"),Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")],
                    'changes':[Change(kind="install-snap", snap_types=["app"])]}
        assert expected == cov

    def test_feat_sys_suite(self):
        data = {'timestamp1': {'system': {'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap routine file-access")],
                         endpoints=[Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]),
                         ]}}}
        retriever = DictRetriever(data)

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', True, suite='suite1')
        expected = {'cmds':[Cmd(cmd="snap list")],
                    'endpoints':[Endpoint(method="GET", path="/v2/snaps")],
                    'changes':[Change(kind="install-snap", snap_types=["app"])]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', True, suite='suite2')
        expected = {'cmds':[Cmd(cmd="snap pack"), Cmd(cmd="snap debug api")],
                    'endpoints':[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', False, suite='suite2')
        expected = {'cmds':[Cmd(cmd="snap pack"), Cmd(cmd="snap debug api"), Cmd(cmd="snap routine file-access")],
                    'endpoints':[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove"),Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', False, suite='doesnotexist')
        assert {} == cov

    def test_feat_sys_task(self):
        data = {'timestamp1': {'system': {'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap routine file-access")],
                         endpoints=[Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]),
                         ]}}}
        retriever = DictRetriever(data)

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', True, task='task1')
        expected = {'cmds':[Cmd(cmd="snap list")],
                    'endpoints':[Endpoint(method="GET", path="/v2/snaps")],
                    'changes':[Change(kind="install-snap", snap_types=["app"])]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', False, task='task1')
        expected = {'cmds':[Cmd(cmd="snap list"),Cmd(cmd="snap routine file-access")],
                    'endpoints':[Endpoint(method="GET", path="/v2/snaps"), Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")],
                    'changes':[Change(kind="install-snap", snap_types=["app"])]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', False, task='doesnotexist')
        assert {} == cov

    def test_feat_sys_variant(self):
        data = {'timestamp1': {'system': {'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap routine file-access")],
                         endpoints=[Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]),
                         ]}}}
        retriever = DictRetriever(data)

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', True, variant='v1')
        expected = {'cmds':[Cmd(cmd="snap pack"),Cmd(cmd="snap debug api")],
                    'endpoints':[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', False, variant='v1')
        expected = {'cmds':[Cmd(cmd="snap pack"),Cmd(cmd="snap debug api"),Cmd(cmd="snap routine file-access")],
                    'endpoints':[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove"),Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]}
        assert expected == cov

        cov = query_features.feat_sys(retriever, 'timestamp1', 'system', False, variant='doesnotexist')
        assert {} == cov

    def test_diff_feat_all(self):
        data = {'timestamp1': {'system': {'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap routine file-access")],
                         endpoints=[Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]),
                         ]}}}
        all_features = {'timestamp1': {
            'cmds':[cmd for entry in data['timestamp1']['system']['tests'] if 'cmds' in entry for cmd in entry['cmds']],
            'endpoints':[endpt for entry in data['timestamp1']['system']['tests'] if 'endpoints' in entry for endpt in entry['endpoints']],
            'changes':[Change(kind="install-snap"),Change(kind="create-recovery-system")],
            'ensures':[Ensure(manager="SnapManager",function="myFunction")],
            'interfaces':[Interface(name="iface")],
            'tasks':[Task(kind="refresh", last_status="Done")]}}
        
        retriever = DictRetriever(data, all_features)

        diff = query_features.diff_all_features(retriever, 'timestamp1', 'system', False)
        expected = {
            'changes':[Change(kind="create-recovery-system")], 
            'ensures':all_features['timestamp1']['ensures'], 
            'interfaces': all_features['timestamp1']['interfaces'],
            'tasks': all_features['timestamp1']['tasks']}
        assert expected == diff

        diff = query_features.diff_all_features(retriever, 'timestamp1', 'system', True)
        expected = {
            'cmds':[Cmd(cmd="snap routine file-access")],
            'endpoints':[Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")],
            'changes':[Change(kind="create-recovery-system")],
            'ensures':all_features['timestamp1']['ensures'], 
            'interfaces': all_features['timestamp1']['interfaces'],
            'tasks': all_features['timestamp1']['tasks']}
        assert expected == diff

    def test_feat_find(self):
        data = {'timestamp1': {'system': {'system':'system', 'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")],
                         tasks=[Task(kind="a",last_status="Done",snap_types=["app"])],
                         ensures=[Ensure(manager="a",function="c")]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),Cmd(cmd="snap debug api")],
                         interfaces=[Interface(name="i",plug_snap_type="snapd",slot_snap_type="app")],
                         endpoints=[Endpoint(method="POST", path="/v", action="b")],
                         changes=[Change(kind="a",snap_types=["app","snapd"])]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap list")],
                         endpoints=[Endpoint(method="POST", path="/v", action="a")],
                         ensures=[Ensure(manager="a",function="b")]),
                         ]}}}
        
        retriever = DictRetriever(data)
        tests = query_features.find_feat(retriever, 'timestamp1', Cmd(cmd="snap list"), False)
        expected = {'system': [TaskIdVariant(suite='suite1',task_name='task1',variant=''),TaskIdVariant(suite='suite2',task_name='task1',variant='v1')]}
        assert json.dumps(expected, default=lambda x: str(x)) == json.dumps(tests, default=lambda x: str(x))

        tests = query_features.find_feat(retriever, 'timestamp1', Cmd(cmd="snap list"), True)
        expected = {'system': [TaskIdVariant(suite='suite1',task_name='task1',variant='')]}
        assert json.dumps(expected, default=lambda x: str(x)) == json.dumps(tests, default=lambda x: str(x))

        tests = query_features.find_feat(retriever, 'timestamp1', Task(kind='a',last_status='Done'), False)
        expected = {'system': [TaskIdVariant(suite='suite1',task_name='task1',variant='')]}
        assert json.dumps(expected, default=lambda x: str(x)) == json.dumps(tests, default=lambda x: str(x))

        tests = query_features.find_feat(retriever, 'timestamp1', Interface(name='i'), False)
        expected = {'system': [TaskIdVariant(suite='suite2',task_name='task2',variant='v1')]}
        assert json.dumps(expected, default=lambda x: str(x)) == json.dumps(tests, default=lambda x: str(x))

        tests = query_features.find_feat(retriever, 'timestamp1',Endpoint(method="POST", path="/v", action="a"), False)
        expected = {'system': [TaskIdVariant(suite='suite2',task_name='task1',variant='v1')]}
        assert json.dumps(expected, default=lambda x: str(x)) == json.dumps(tests, default=lambda x: str(x))

        tests = query_features.find_feat(retriever, 'timestamp1',Ensure(manager="a",function="b"), False)
        expected = {'system': [TaskIdVariant(suite='suite2',task_name='task1',variant='v1')]}
        assert json.dumps(expected, default=lambda x: str(x)) == json.dumps(tests, default=lambda x: str(x))

        tests = query_features.find_feat(retriever, 'timestamp1',Change(kind="a"), False)
        expected = {'system': [TaskIdVariant(suite='suite2',task_name='task2',variant='v1')]}
        assert json.dumps(expected, default=lambda x: str(x)) == json.dumps(tests, default=lambda x: str(x))


    def test_task_list(self):
        data = {'timestamp1': {'system': {'system':'system', 'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),Cmd(cmd="snap debug api")]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap list")]),
                         ]}}}
        
        retriever = DictRetriever(data)
        tasks = query_features.task_list(retriever, 'timestamp1')
        assert len(tasks) == 3
        assert TaskIdVariant(suite='suite1',task_name='task1',variant='') in tasks
        assert TaskIdVariant(suite='suite2',task_name='task2',variant='v1') in tasks
        assert TaskIdVariant(suite='suite2',task_name='task1',variant='v1') in tasks


    def test_sys_caching(self):
        data = {'timestamp1': {
                'system1': {'system':'system1', 'tests': [
                    TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                                cmds=[Cmd(cmd="some command")])]},
                'system2': {'system':'system2', 'tests': [
                    TaskFeatures(suite='suite2', task_name='task2', success=True, variant='',
                                cmds=[Cmd(cmd="system2")]),]}},}
        retriever = DictRetriever(data)
        retriever.get_single_json('timestamp1', 'system1')
        assert len(retriever.cache) == 0
        res = retriever.get_systems('timestamp1', ['system1'])
        assert res[0]['tests'][0]['cmds'][0]['cmd'] == 'some command'

        # Before data is cached, it's possible to overwrite values
        retriever.data['timestamp1']['system1']['tests'][0]['cmds'][0]['cmd'] = 'system1'
        res = retriever.get_systems('timestamp1', ['system1'])
        assert res[0]['tests'][0]['cmds'][0]['cmd'] == 'system1'
        assert len(retriever.cache) == 0

        # By calling get_systems without specifying a system list, the base
        # class will cache the result
        retriever.get_systems('timestamp1')
        assert len(retriever.cache) == 1
        assert 'timestamp1' in retriever.cache
        assert 'systems' in retriever.cache['timestamp1']

        # To prove the cached data is being used, here we change the underlying data
        retriever.data['timestamp1']['system1']['tests'][0]['cmds'][0]['cmd'] = 'new'
        res = retriever.get_single_json('timestamp1', 'system1')
        assert res['tests'][0]['cmds'][0]['cmd'] == 'system1'

        # get_systems and get_single_json all use the cached data
        res = retriever.get_systems('timestamp1', ['system1'])
        assert res[0]['tests'][0]['cmds'][0]['cmd'] == 'system1'
        res = retriever.get_systems('timestamp1')
        assert res[0]['tests'][0]['cmds'][0]['cmd'] == 'system1'
        res = retriever.get_single_json('timestamp1', 'system1')
        assert res['tests'][0]['cmds'][0]['cmd'] == 'system1'


    def test_all_features_caching(self):
        data = {'timestamp1': {'system': {'tests': [
            TaskFeatures(suite='suite1', task_name='task1', success=True, variant='',
                         cmds=[Cmd(cmd="snap list")],
                         endpoints=[
                             Endpoint(method="GET", path="/v2/snaps")],
                         changes=[Change(kind="install-snap", snap_types=["app"])]),
            TaskFeatures(suite='suite2', task_name='task2', success=True, variant='v1',
                         cmds=[Cmd(cmd="snap pack"),
                               Cmd(cmd="snap debug api")],
                         endpoints=[Endpoint(method="POST", path="/v2/snaps/{name}", action="remove")]),
            TaskFeatures(suite='suite2', task_name='task1', success=False, variant='v1',
                         cmds=[Cmd(cmd="snap routine file-access")],
                         endpoints=[Endpoint(method="POST", path="/v2/system-info", action="advise-system-key-mismatch")]),
                         ]}}}
        all_features = {'timestamp1': {
            'cmds':[cmd for entry in data['timestamp1']['system']['tests'] if 'cmds' in entry for cmd in entry['cmds']],
            'endpoints':[endpt for entry in data['timestamp1']['system']['tests'] if 'endpoints' in entry for endpt in entry['endpoints']],
            'changes':[Change(kind="install-snap"),Change(kind="create-recovery-system")],
            'ensures':[Ensure(manager="SnapManager",function="myFunction")],
            'interfaces':[Interface(name="iface")],
            'tasks':[Task(kind="refresh", last_status="Done")]}}
        
        retriever = DictRetriever(data, all_features)
        
        assert len(retriever.cache) == 0
        retriever.get_all_features('timestamp1')
        assert len(retriever.cache) == 1
        assert 'timestamp1' in retriever.cache
        assert 'all-features' in retriever.cache['timestamp1']

        # all-features is now cached. To prove the result
        # comes from the cache, here we overwrite some data
        # and see the original values reflected in the output
        retriever.all_features['cmds'] = []
        res = retriever.get_all_features('timestamp1')
        assert len(res['cmds']) > 0



    @patch('argparse.ArgumentParser.parse_args')
    def test_dirretriever_list(self, parse_args_mock: Mock):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system1'},
            {'timestamp': '2025-05-04', 'system': 'system2'},
            {'timestamp': '2025-05-05', 'system': 'system2'}
        ]
        with DirMocker(data, do_patch_stdout=True) as dm:
            parse_args_mock.return_value = argparse.Namespace(
                command='list',
                file=None,
                dir=dm.get_dir()
            )
            query_features.main()
            actual = json.loads(dm.get_stdout())
            assert 2 == len(actual)
            in_actual = {'timestamp': '2025-05-04', 'systems': ['system1', 'system2']} in actual \
                or {'timestamp': '2025-05-04', 'systems': ['system2', 'system1']} in actual
            assert in_actual
            assert {'timestamp': '2025-05-05','systems': ['system2']} in actual


    @patch('argparse.ArgumentParser.parse_args')
    def test_mongoretriever_list(self, parse_args_mock: Mock):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system1'},
            {'timestamp': '2025-05-04', 'system': 'system2'},
            {'timestamp': '2025-05-05', 'system': 'system2'}
        ]
        with MongoMocker(data, do_patch_stdout=True) as mm:
            parse_args_mock.return_value = argparse.Namespace(
                command='list',
                file=StringIO(''),
                dir=None
            )
            query_features.main()
            actual = json.loads(mm.get_stdout())
            assert 2, len(actual)
            in_actual = {'timestamp': '2025-05-04T00:00:00', 'systems': ['system1', 'system2']} in actual \
                or {'timestamp': '2025-05-04T00:00:00', 'systems': ['system2', 'system1']} in actual
            assert in_actual
            assert {'timestamp': '2025-05-05T00:00:00', 'systems': ['system2']} in actual


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_diff_systems(self, parse_args_mock: Mock, mocker_class: str):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system', 'tests': [
                {'cmds': [{'cmd': 'a'}, {'cmd': 'b'}], 'endpoints': [{'1': 'a'}]},
                {'cmds': [{'cmd': 'd'}], 'endpoints': [{'5': 'd'}]},
            ]},
            {'timestamp': '2025-05-05', 'system': 'system', 'tests': [
                {'cmds': [{'cmd': 'a'}, {'cmd': 'c'}], 'endpoints': [{'1': 'b'}, {'2': 'a'}]},
                {'cmds': [{'cmd': 'd'}], 'tasks': [{'task': 'a'}]},
                {'cmds': [{'cmd': 'e'}], 'endpoints': [{'5': 'd'}]}
            ]}
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data, do_patch_stdout=True) as mocker:
            parse_args_mock.return_value = argparse.Namespace(
                command='diff',
                diff_cmd='systems',
                file=StringIO('') if mocker_class == "MongoMocker" else None,
                dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                timestamp1='2025-05-04',
                system1='system',
                timestamp2='2025-05-05',
                system2='system',
                remove_failed=False,
                only_same=False,
                match_snap_types=True,
                sort_by_test=False
            )
            query_features.main()
            expected = {'cmds': [{'cmd': 'b'}], 'endpoints': [{'1': 'a'}]}
            actual = json.loads(mocker.get_stdout())
            assert expected == actual


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_diff_all(self, parse_args_mock: Mock, mocker_class: str):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system', 'tests': [
                {'cmds': [Cmd(cmd='a'), Cmd(cmd='b')], 'endpoints': [Endpoint(method='a',path='/a')]},
                {'cmds': [Cmd(cmd='d')], 'endpoints': [Endpoint(method='b', path='/b', action='b')]},
                {'tasks': [Task(kind='a',last_status='Done',snap_types=['a'])], 'changes': [Change(kind='change')],},
                {'interfaces': [Interface(name='iface')],'ensures': [Ensure(manager='mgr',function='func')]}
            ]},
            {'timestamp': '2025-05-05', 'system': 'system', 'tests': [
                {'cmds': [Cmd(cmd='a'), Cmd(cmd='c')], 'endpoints': [Endpoint(method='c',path='/c'),Endpoint(method='d',path='/d',action='d')]},
                {'cmds': [Cmd(cmd='d')], 'tasks': [Task(kind='a',last_status='Done',snap_types=['a'])]},
                {'cmds': [Cmd(cmd='e')], 'endpoints': [Endpoint(method='c',path='/c')]}
            ]},
            {'timestamp': '2025-05-04', 'all_features': True,
                'cmds': [Cmd(cmd='a'), Cmd(cmd='b'), Cmd(cmd='c'), Cmd(cmd='d'), Cmd(cmd='e'), Cmd(cmd='f')], 
                'endpoints': [Endpoint(method='a',path='/a'),
                              Endpoint(method='b', path='/b', action='b'),
                              Endpoint(method='c',path='/c'),
                              Endpoint(method='d',path='/d',action='d')],
                'tasks': [Task(kind='a',last_status='Done'),Task(kind='a',last_status='Error')],
                'interfaces': [Interface(name='iface')],
                'changes': [Change(kind='change')],
                'ensures': [Ensure(manager='mgr',function='func')]
            }
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data, do_patch_stdout=True) as mocker:
            parse_args_mock.return_value = argparse.Namespace(
                command='diff',
                diff_cmd='all-features',
                file=StringIO('') if mocker_class == "MongoMocker" else None,
                dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                timestamp='2025-05-04',
                system='system',
                remove_failed=False
            )
            query_features.main()
            expected = {'cmds': [Cmd(cmd='c'),Cmd(cmd='e'),Cmd(cmd='f')],
                        'endpoints': [Endpoint(method='c',path='/c'),Endpoint(method='d',path='/d',action='d')],
                        'tasks': [Task(kind='a',last_status='Error')]}
            actual = json.loads(mocker.get_stdout())
            assert expected == actual


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_dup(self, parse_args_mock: Mock, mocker_class: str):
        data = [{'timestamp': '2025-05-04', 'system': 'system', 'tests': [
            TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'a'}, {'cmd': 'b'}], endpoints=[{'1': 'a'}]),
            TaskFeatures(task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'5': 'd'}]),
            TaskFeatures(task_name='task3', suite='suite2', variant='', cmds=[{'cmd': 'd'}]),
            TaskFeatures(task_name='task4', suite='suite1', variant='v1', endpoints=[{'1': 'a'}])
        ]}
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data, do_patch_stdout=True) as mocker:
            parse_args_mock.return_value = argparse.Namespace(
                command='dup',
                file=StringIO('') if mocker_class == "MongoMocker" else None,
                dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                timestamp='2025-05-04',
                system='system',
                remove_failed=False,
            )
            query_features.main()
            actual = json.loads(mocker.get_stdout())
            assert 2 == len(actual)
            assert 'suite2:task3' in actual
            assert 'suite1:task4:v1' in actual


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_export(self, parse_args_mock: Mock, mocker_class: str):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system1', 'tests': [
                TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'a'}, {'cmd': 'b'}], endpoints=[{'1': 'a'}]),
                TaskFeatures(task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'5': 'd'}])
            ]},
            {'timestamp': '2025-05-05', 'system': 'system2', 'tests': [
                TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'c'}, {'cmd': 'd'}], endpoints=[{'1': 'a'}]),
                TaskFeatures(task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'2': 'q'}])
            ]},
            {'timestamp': '2025-05-06', 'system': 'system3', 'tests': [
                TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'a'}])
            ]},
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data) as mocker:
            with tempfile.TemporaryDirectory() as tmpdir:
                parse_args_mock.return_value = argparse.Namespace(
                    command='export',
                    file=StringIO('') if mocker_class == "MongoMocker" else None,
                    dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                    timestamps=['2025-05-04', '2025-05-05'],
                    systems=None,
                    output=tmpdir,
                )
                with patch('sys.stderr', new=StringIO()) as stderr_patch:
                    query_features.main()
                    assert stderr_patch.getvalue().startswith('could not find all features at timestamp 2025-05-04')

                assert os.path.isdir(os.path.join(tmpdir, '2025-05-04'))
                assert os.path.isdir(os.path.join(tmpdir, '2025-05-05'))
                assert os.path.isfile(os.path.join(tmpdir, '2025-05-04', 'system1.json'))
                assert os.path.isfile(os.path.join(tmpdir, '2025-05-05', 'system2.json'))


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_export_with_all(self, parse_args_mock: Mock, mocker_class: str):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system1', 'tests': [
                TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'a'}, {'cmd': 'b'}], endpoints=[{'1': 'a'}]),
                TaskFeatures(task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'5': 'd'}])
            ]},
            {'timestamp': '2025-05-05', 'system': 'system2', 'tests': [
                TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'c'}, {'cmd': 'd'}], endpoints=[{'1': 'a'}]),
                TaskFeatures(task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'2': 'q'}])
            ]},
            {'timestamp': '2025-05-06', 'system': 'system3', 'tests': [
                TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'a'}])
            ]},
            {'timestamp': '2025-05-04', 'all_features': True, 'cmds': [{'cmd': 'a'},{'cmd': 'b'},{'cmd': 'c'},{'cmd': 'd'}]},
            {'timestamp': '2025-05-05', 'all_features': True, 'cmds': [{'cmd': 'a'},{'cmd': 'b'},{'cmd': 'c'},{'cmd': 'd'}]},
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data) as mocker:
            with tempfile.TemporaryDirectory() as tmpdir:
                parse_args_mock.return_value = argparse.Namespace(
                    command='export',
                    file=StringIO('') if mocker_class == "MongoMocker" else None,
                    dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                    timestamps=['2025-05-04', '2025-05-05'],
                    systems=None,
                    output=tmpdir,
                )
                query_features.main()

                assert os.path.isdir(os.path.join(tmpdir, '2025-05-04'))
                assert os.path.isdir(os.path.join(tmpdir, '2025-05-05'))
                assert os.path.isfile(os.path.join(tmpdir, '2025-05-04', 'system1.json'))
                assert os.path.isfile(os.path.join(tmpdir, '2025-05-05', 'system2.json'))
                assert os.path.isfile(os.path.join(tmpdir, '2025-05-04', 'all-features.json'))
                assert os.path.isfile(os.path.join(tmpdir, '2025-05-05', 'all-features.json'))


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_feat_sys(self, parse_args_mock: Mock, mocker_class: str):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system1', 'tests': [
                TaskFeatures(task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'a'}, {'cmd': 'b'}], endpoints=[{'1': 'a'}], success=True),
                TaskFeatures(task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'5': 'd'}], success=True),
                TaskFeatures(task_name='task3', suite='suite2', variant='', cmds=[{'cmd': 'c'}], success=False)
            ]}
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data, do_patch_stdout=True) as mocker:
            parse_args_mock.return_value = argparse.Namespace(
                command='feat',
                features_cmd='sys',
                file=StringIO('') if mocker_class == "MongoMocker" else None,
                dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                timestamp='2025-05-04',
                system='system1',
                suite=None,
                task=None,
                variant=None,
                remove_failed=True
            )
            query_features.main()

            output = json.loads(mocker.get_stdout())
            expected = {'cmds':[{'cmd':'a'},{'cmd':'b'},{'cmd':'d'}],
                        'endpoints':[{'1':'a'},{'5':'d'}]}
            assert expected == output


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_feat_all(self, parse_args_mock: Mock, mocker_class: str):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system', 'tests': [
                {'cmds': [{'cmd': 'a'}, {'cmd': 'b'}], 'endpoints': [{'1': 'a'}]},
                {'cmds': [{'cmd': 'd'}], 'endpoints': [{'5': 'd'}]},
            ]},
            {'timestamp': '2025-05-04', 'all_features': True,
                'cmds': [{'cmd': 'a'}, {'cmd': 'b'}, {'cmd': 'c'}, {'cmd': 'd'}, {'cmd': 'e'}, {'cmd': 'f'}], 
                'endpoints': [{'1': 'a'},{'1': 'b'},{'2': 'a'},{'5': 'd'}]
            }
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data, do_patch_stdout=True) as mocker:
            parse_args_mock.return_value = argparse.Namespace(
                command='feat',
                features_cmd='all',
                file=StringIO('') if mocker_class == "MongoMocker" else None,
                dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                timestamp='2025-05-04',
                system='system1',
                remove_failed=True
            )
            query_features.main()

            output = json.loads(mocker.get_stdout())
            expected = {'cmds':data[1]['cmds'],
                        'endpoints':data[1]['endpoints']}
            assert expected == output


    @pytest.mark.parametrize("mocker_class", ["MongoMocker","DirMocker"])
    @patch('argparse.ArgumentParser.parse_args')
    def test_retriever_feat_find(self, parse_args_mock: Mock, mocker_class: str):
        data = [
            {'timestamp': '2025-05-04', 'system': 'system1', 'tests': [
                TaskFeatures(success=True, task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'a'}, {'cmd': 'b'}], endpoints=[{'1': 'a'}]),
                TaskFeatures(success=True, task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'2': 'q'}])
            ]},
            {'timestamp': '2025-05-04', 'system': 'system2', 'tests': [
                TaskFeatures(success=True, task_name='task1', suite='suite1', variant='', cmds=[{'cmd': 'c'}, {'cmd': 'd'}], endpoints=[{'1': 'a'}]),
                TaskFeatures(success=False, task_name='task2', suite='suite1', variant='', cmds=[{'cmd': 'd'}], endpoints=[{'2': 'q'}])
            ]},
            {'timestamp': '2025-05-04', 'all_features': True,
                'cmds': [{'cmd': 'a'}, {'cmd': 'b'}, {'cmd': 'c'}, {'cmd': 'd'}, {'cmd': 'e'}, {'cmd': 'f'}], 
                'endpoints': [{'1': 'a'},{'1': 'b'},{'2': 'a'},{'5': 'd'}]
            }
        ]
        Mocker = globals()[mocker_class]
        with Mocker(data, do_patch_stdout=True) as mocker:
            parse_args_mock.return_value = argparse.Namespace(
                command='feat',
                features_cmd='find',
                file=StringIO('') if mocker_class == "MongoMocker" else None,
                dir=mocker.get_dir() if mocker_class == "DirMocker" else None,
                timestamp='2025-05-04',
                feat='{"cmd":"d"}',
                system=None,
                remove_failed=True,
                match_snap_types=True
            )
            query_features.main()

            output = json.loads(mocker.get_stdout())
            expected = {'system1':[str(TaskIdVariant(suite='suite1',task_name='task2',variant=''))],
                        'system2':[str(TaskIdVariant(suite='suite1',task_name='task1',variant=''))]}
            assert expected == output



#!/usr/bin/env python3

from __future__ import annotations

import argparse
from typing import IO

from common import LockOpTrace, get_next_match


class LockOpTraceFileReader:
    """
    LocksFileReader: Reads the traces file and retrieves differences
    with a second traces file
    """

    lines: dict[LockOpTrace, LockOpTrace]

    def __init__(self, traces_file: IO[str]):
        self.traces = {}
        self.lines = []

        self._read(traces_file)

    def _read(self, traces_file: IO[str]) -> None:
        self.lines = traces_file.readlines()

        current_line = 0
        # Read the tests
        while current_line < len(self.lines):
            trace_lines = self._current_trace(current_line)
            if len(trace_lines) == 0:
                raise RuntimeError("Error parsing traces file")

            # Remove empty lines and title
            cleaned_lines = [
                line for line in trace_lines if line.strip() and
                "TRACE" not in line
            ]

            # Remove the file and line numbers
            # Line numbers could easily change over time
            # File paths could include snap versions which used to change
            function_lines = [line.split(" ", 1)[1] for line in cleaned_lines]

            # Save the traces in a dict to be able to print the version of the trace
            # with the file and line number which is not used to compare
            self.traces[LockOpTrace(function_lines)] = LockOpTrace(cleaned_lines)
            current_line = current_line + len(trace_lines)

    # This function works to detect the current trace
    def _current_trace(self, start_line: int) -> list[str]:
        next_match = get_next_match(self.lines, start_line, "---")
        if next_match == -1:
            return self.lines[start_line:len(self.lines)]
        return self.lines[start_line:next_match]

    # Retieve the traces which are in other and are not in self
    def get_diff(self, other: LockOpTraceFileReader) -> list[LockOpTrace]:
        diff_traces = []
        my_traces = self.traces.keys()
        for key_trace, val_trace in other.traces.items():
            if key_trace not in my_traces:
                diff_traces.append(val_trace)
        return diff_traces

    # Print the traces which are in other and are not in self
    def print_diff(self, other: LockOpTraceFileReader):
        for trace in self.get_diff(other):
            trace.print()


"""
A traces file is a sequence of traces separated by an empty line

It parses the traces file and will print the traces included in the
locks-file which are not present in the baseline.
"""


def _make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="state-locks-filter",
        description="Reads a locks file and extracts lock information.",
    )
    parser.add_argument(
        "-b",
        "--baseline-file",
        type=argparse.FileType("r"),
        required=True,
        help="Baseline traces file",
    )
    parser.add_argument(
        "-f",
        "--locks-file",
        type=argparse.FileType("r"),
        required=True,
        help="Test traces file",
    )

    return parser


if __name__ == "__main__":
    parser = _make_parser()
    args = parser.parse_args()

    locks_reader_baseline = LockOpTraceFileReader(args.baseline_file)
    locks_reader_locks = LockOpTraceFileReader(args.locks_file)
    locks_reader_baseline.print_diff(locks_reader_locks)



import unittest
from io import StringIO
from traces_diff import LockOpTraceFileReader


class TestLockOpTraceFileReader(unittest.TestCase):
    def setUp(self):
        self.baseline_trace = """---TRACE 1---\n\nfile1:123 lock A\nfile1:234 unlock A\n"""
        # Diff
        self.sample_trace_1 = """---TRACE 2---\n\nfile2:123 lock B\nfile1:123 unlock A\n"""
        self.sample_trace_2 = """---TRACE 2---\n\nfile1:123 lock B\nfile1:234 unlock A\n"""
        # No Diff
        self.sample_trace_3 = """---TRACE 3---\n\nfile1:123 lock A\nfile1:234 unlock A\n"""
        self.sample_trace_4 = """---TRACE 4---\n\nfile1:123 lock A\nfile1:235 unlock A\n"""
        self.sample_trace_5 = """---TRACE 5---\n\nfile2:123 lock A\nfile1:234 unlock A\n"""

    def test_trace_diff_1(self):
        reader1 = LockOpTraceFileReader(StringIO(self.baseline_trace))
        reader2 = LockOpTraceFileReader(StringIO(self.sample_trace_1))
        self.assertIsNot(reader1.get_diff(reader2), [])

    def test_trace_diff_2(self):
        reader1 = LockOpTraceFileReader(StringIO(self.baseline_trace))
        reader2 = LockOpTraceFileReader(StringIO(self.sample_trace_2))
        self.assertIsNot(reader1.get_diff(reader2), [])

    def test_trace_no_diff_1(self):
        reader1 = LockOpTraceFileReader(StringIO(self.baseline_trace))
        reader2 = LockOpTraceFileReader(StringIO(self.sample_trace_3))
        self.assertIs(len(reader1.get_diff(reader2)), 0)

    def test_trace_no_diff_2(self):
        reader1 = LockOpTraceFileReader(StringIO(self.baseline_trace))
        reader2 = LockOpTraceFileReader(StringIO(self.sample_trace_4))
        self.assertIs(len(reader1.get_diff(reader2)), 0)

    def test_trace_no_diff_3(self):
        reader1 = LockOpTraceFileReader(StringIO(self.baseline_trace))
        reader2 = LockOpTraceFileReader(StringIO(self.sample_trace_5))
        self.assertIs(len(reader1.get_diff(reader2)), 0)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
"""
Pure-python implementation of the status of the stax() system call.
"""
from ctypes import CDLL, c_char, c_long, create_string_buffer, get_errno
from ctypes.util import find_library
from enum import Enum
from errno import ENOSYS, EPERM
from os import uname


__all__ = ("SyscallStatus", "evaluate_statx_support")


class SyscallStatus(Enum):
    """Syscall status encodes the status of a system call."""

    SUPPORTED = "supported"
    MISSING = "missing"
    BLOCKED = "blocked"


def evaluate_statx_support() -> SyscallStatus:
    """
    Evaluate status of the statx system call.

    The statx system call was introduced in the Linux kernel 4.11.  A snap
    application running under seccomp confinement may experience one of three
    results when accessing statx: Supported, namely implemented by the kernel
    and allowed by the seccomp filter. Missing, namely not implemented by the
    kernel but not blocked by seccomp. Blocked namely not allowed by seccomp
    and either implemented or not by the kernel.
    """
    machine = uname()[4]
    sys_statx_table = {
        "i686": 383,
        "i386": 383,
        "x86_64": 332,
        "arm": 397,
        "aarch64": 291,
        "ppc": 383,
        "ppcel64": 383,
        "s390x": 379,
    }
    try:
        SYS_STATX = sys_statx_table[machine]
    except KeyError:
        raise Exception("unsupported architecture {}".format(machine))
    libc_name = find_library("c")
    if libc_name is None:
        raise Exception("cannot find the C library")
    libc = CDLL(libc_name, use_errno=True)
    syscall = libc.syscall
    buf_4k = c_char * 4096
    dummy_buf = buf_4k()
    empty_string = create_string_buffer(b"")
    AT_FDCWD = -100
    AT_EMPTY_PATH = 0x1000
    # statx(2) prototype:
    #
    # int statx(int dirfd, const char *pathname, int flags,
    #           unsigned int mask, struct statx *statxbuf);
    #
    # We don't attempt to define struct statx here, instead we just
    # provide a buffer with ample space, so that the system call has sufficient
    # space to write the result to.
    #
    # Python equivalent of:
    #
    #   char dummy_buf[4096];
    #   syscall(SYS_STATX, AT_FDCWD, "", AT_EMPTY_PATH, 0, dummy_buf);
    retval = syscall(
        c_long(SYS_STATX),
        c_long(AT_FDCWD),
        empty_string,
        c_long(AT_EMPTY_PATH),
        c_long(0),
        dummy_buf,
    )
    errno = get_errno()
    if retval == 0:
        return SyscallStatus.SUPPORTED
    elif retval == -1 and errno == ENOSYS:
        return SyscallStatus.MISSING
    elif retval == -1 and errno == EPERM:
        return SyscallStatus.BLOCKED
    raise Exception(
        "unexpected result of statx, retval: {}, errno: {}".format(retval, errno)
    )


def main() -> None:
    try:
        print("statx: {}".format(evaluate_statx_support().value))
    except Exception as exc:
        print("statx: error: {}".format(exc))


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

from dash import html, dcc, dash_table
import dash_bootstrap_components as dbc
import dash_daq as daq
import os
import sys

# To ensure this can be run from any point in the filesystem,
# add parent folder to path to permit relative imports
sys.path.append(os.path.dirname(os.path.abspath(__file__)))

from dashboard_callbacks import *

app.layout = html.Div(
    style={
        "padding": "15px",
    },
    children=[
        html.Div(children="", style={"fontSize": "24px", "marginBottom": "10px"}),
        html.H4(
            "1. First input path for either a mongo credential file or a folder containing the data."
        ),
        dcc.Loading(
            html.Div(
                [
                    dcc.Input(
                        id="input-path",
                        type="text",
                        placeholder="full path to mongo cred file or to folder with data.",
                    ),
                    dbc.Button("Set data", id="set-data-button"),
                    html.Div(id="empty-div", style={"display": "none"}),
                ]
            ),
        ),
        dcc.ConfirmDialog(
            id="error-with-retriever",
            message="The indicated path does not exist",
            displayed=False,
        ),
        html.Div(children="", style={"fontSize": "24px", "marginBottom": "20px"}),
        html.H4("2. Explore data"),
        html.Div(children="", style={"fontSize": "24px", "marginBottom": "20px"}),
        dbc.Button(
            "Systems coverage vs. all features",
            id={"type": "toggle-button", "index": 1},
            n_clicks=0,
            className="mb-3",
            disabled=True,
        ),
        dbc.Collapse(
            html.Div(
                [
                    html.H4(
                        "Calculates the difference between all features and each system's features"
                    ),
                    html.H5("Calculated as follows:"),
                    html.H6(
                        "commands, ensures, endpoints: the entire feature data is used for equality check"
                    ),
                    html.H6(
                        "tasks: only kind and last_status are used for equality check"
                    ),
                    html.H6("changes: only kind is used for equality check"),
                    html.H6("interfaces: only name is used for equality check"),
                    dcc.Dropdown(
                        id={"type": "timestamp-dropdown", "index": 1},
                        options=timestamp_options,
                        placeholder="Select timestamp",
                    ),
                    dcc.Loading(
                        html.Div(
                            [
                                dash_table.DataTable(
                                    id="all-features-diff-table",
                                    columns=[],
                                    data=[],
                                    filter_action="native",
                                    sort_action="native",
                                    active_cell=None,
                                    style_cell={
                                        "textAlign": "center",
                                        "minWidth": "100px",
                                        "maxWidth": "200px",
                                        "whiteSpace": "normal",
                                    },
                                    style_table={
                                        "overflowX": "auto",
                                        "maxWidth": "900px",
                                        "margin": "auto",
                                    },
                                ),
                                html.Div(
                                    id="all-features-diff-cell-data-container",
                                    style={
                                        "display": "inline-block",
                                        "marginLeft": "20px",
                                        "verticalAlign": "top",
                                        "flex": 1,
                                        "overflow": "auto",
                                        "maxWidth": "900px",
                                    },
                                ),
                            ],
                            id="all-features-diff-container",
                            style={
                                "display": "flex",
                                "justifyContent": "center",
                                "alignItems": "flex-start",
                                "gap": "20px",
                                "maxWidth": "1900px",
                                "margin": "auto",
                            },
                        ),
                    ),
                ]
            ),
            id={"type": "collapse", "index": 1},
            is_open=False,
        ),
        html.Div(children="", style={"fontSize": "24px", "marginBottom": "20px"}),
        dbc.Button(
            "Compare feature coverage between systems",
            id={"type": "toggle-button", "index": 2},
            n_clicks=0,
            className="mb-3",
            disabled=True,
        ),
        dbc.Collapse(
            html.Div(
                [
                    html.Div(
                        children="Calculates difference in coverage between set(system_1_features) - set(system_2_features)",
                        style={"fontSize": "20px", "marginBottom": "20px"},
                    ),
                    html.Div(
                        [
                            html.Div(
                                children="System 1", style={"marginBottom": "20px"}
                            ),
                            dcc.Dropdown(
                                id={"type": "timestamp-dropdown", "index": 2},
                                options=timestamp_options,
                                placeholder="Select timestamp",
                            ),
                            dcc.Dropdown(
                                id={"type": "systems-dropdown", "index": 2},
                                options=[],
                                placeholder="Select system",
                            ),
                            dcc.Input(
                                id={"type": "systems-regex-input", "index": 2},
                                type="text",
                                placeholder="Use regex here to select systems instead of dropdown",
                                debounce=True,
                            ),
                            html.Div(children="", style={"marginBottom": "10px"}),
                            html.Button(
                                "🔄",
                                id="switch-button",
                                n_clicks=0,
                                style={
                                    "fontSize": "20px",
                                    "padding": "4px 8px",
                                    "margin": "0 10px",
                                    "borderRadius": "50%",
                                    "border": "1px solid #ccc",
                                    "backgroundColor": "white",
                                    "cursor": "pointer",
                                    "lineHeight": "1",
                                    "display": "inline-flex",
                                    "alignItems": "center",
                                    "justifyContent": "center",
                                    "width": "32px",
                                    "height": "32px",
                                },
                            ),
                            html.Div(
                                children="System 2",
                                style={"marginTop": "10px", "marginBottom": "20px"},
                            ),
                            dcc.Dropdown(
                                id={"type": "timestamp-dropdown", "index": 3},
                                options=timestamp_options,
                                placeholder="Select timestamp",
                            ),
                            dcc.Dropdown(
                                id={"type": "systems-dropdown", "index": 3},
                                options=[],
                                placeholder="Select system",
                            ),
                            daq.BooleanSwitch(
                                id="remove-failed-switch",
                                label="Remove failed tests",
                            ),
                            daq.BooleanSwitch(
                                id="only-same-switch",
                                label="Only compare features across tests that are present in both systems",
                            ),
                            daq.BooleanSwitch(
                                id="match-snap-types",
                                label="Match snap types when comparing features",
                                disabled=False
                            ),
                        ],
                        style={"width": "25%"},
                    ),
                    dcc.Loading(
                        html.Div(id="coverage-diff-container"),
                    ),
                    dbc.Modal(
                        [
                            dbc.ModalHeader(dbc.ModalTitle("Tests with feature")),
                            dbc.ModalBody(id="coverage-diff-modal-body"),
                        ],
                        id="coverage-diff-modal",
                        size="xl",
                        is_open=False,
                    ),
                ]
            ),
            id={"type": "collapse", "index": 2},
            is_open=False,
        ),
        html.Div(children="", style={"fontSize": "24px", "marginBottom": "20px"}),
        dbc.Button(
            "Feature coverage matrix per system",
            id={"type": "toggle-button", "index": 3},
            n_clicks=0,
            className="mb-3",
            disabled=True,
        ),
        dbc.Collapse(
            html.Div(
                [
                    html.Div(
                        children="Calculates the feature coverage matrix for all systems",
                        style={"fontSize": "20px", "marginBottom": "20px"},
                    ),
                    html.Div(children="Timestamp", style={"marginBottom": "20px"}),
                    html.Div(
                        [
                            dcc.Dropdown(
                                id={"type": "timestamp-dropdown", "index": 4},
                                options=timestamp_options,
                                placeholder="Select timestamp",
                            ),
                            html.Div(
                                children="coverage data filter",
                                style={
                                    "fontSize": "16px",
                                    "marginTop": "20px",
                                    "marginBottom": "20px",
                                },
                            ),
                            dcc.Dropdown(
                                id="suite-dropdown-filter",
                                options=[],
                                placeholder="Select suite",
                                clearable=True,
                            ),
                            dcc.Dropdown(
                                id="task-dropdown-filter",
                                options=[],
                                placeholder="Select task",
                                clearable=True,
                            ),
                            dcc.Dropdown(
                                id="variant-dropdown-filter",
                                options=[],
                                placeholder="Select variant",
                                clearable=True,
                            ),
                            daq.BooleanSwitch(
                                id="coverage-remove-failed-switch",
                                label="Remove failed tests",
                            ),
                        ],
                        style={"width": "25%"},
                    ),
                    dcc.Loading(
                        html.Div(
                            [
                                dash_table.DataTable(
                                    id="coverage-matrix-table",
                                    filter_action="native",
                                    sort_action="native",
                                    style_cell={
                                        "textAlign": "center",
                                        "minWidth": "100px",
                                        "maxWidth": "200px",
                                        "whiteSpace": "normal",
                                    },
                                    style_table={
                                        "overflowX": "auto",
                                        "maxWidth": "900px",
                                        "margin": "auto",
                                    },
                                ),
                                html.Div(
                                    id="cell-data-container",
                                    style={
                                        "display": "inline-block",
                                        "marginLeft": "20px",
                                        "verticalAlign": "top",
                                        "flex": 1,
                                        "overflow": "auto",
                                        "maxWidth": "900px",
                                    },
                                ),
                            ],
                            id="coverage-matrix-container",
                            style={
                                "display": "flex",
                                "justifyContent": "center",
                                "alignItems": "flex-start",
                                "gap": "20px",
                                "maxWidth": "1900px",
                                "margin": "auto",
                            },
                        ),
                    ),
                    dbc.Modal(
                        [
                            dbc.ModalHeader(dbc.ModalTitle("Tests with feature")),
                            dbc.ModalBody(id="coverage-modal-body"),
                        ],
                        id="coverage-modal",
                        size="xl",
                        is_open=False,
                    ),
                ]
            ),
            id={"type": "collapse", "index": 3},
            is_open=False,
        ),
        html.Div(children="", style={"fontSize": "24px", "marginBottom": "20px"}),
        dbc.Button(
            "Duplicate features",
            id={"type": "toggle-button", "index": 4},
            n_clicks=0,
            className="mb-3",
            disabled=True,
        ),
        dbc.Collapse(
            html.Div(
                [
                    html.Div(children="Timestamp", style={"marginBottom": "20px"}),
                    dcc.Dropdown(
                        id={"type": "timestamp-dropdown", "index": 5},
                        options=timestamp_options,
                        placeholder="Select timestamp",
                    ),
                    dcc.Dropdown(
                        id={"type": "systems-dropdown", "index": 5},
                        options=[],
                        placeholder="Select system",
                    ),
                    html.Div(
                        [
                            dcc.Loading(
                                dash_table.DataTable(
                                    id="duplicate-table",
                                    filter_action="native",
                                    sort_action="native",
                                    style_cell={
                                        "textAlign": "center",
                                        "minWidth": "100px",
                                        "maxWidth": "200px",
                                        "whiteSpace": "normal",
                                    },
                                    style_table={
                                        "overflowX": "auto",
                                        "maxWidth": "900px",
                                        "margin": "auto",
                                    },
                                ),
                            ),
                            html.Div(
                                id="cell-duplicate-container",
                                style={
                                    "display": "inline-block",
                                    "marginLeft": "20px",
                                    "verticalAlign": "top",
                                    "flex": 1,
                                    "overflow": "auto",
                                    "maxWidth": "900px",
                                },
                            ),
                        ],
                        id="duplicate-matrix-container",
                        style={
                            "display": "flex",
                            "justifyContent": "center",
                            "alignItems": "flex-start",
                            "gap": "20px",
                            "maxWidth": "1900px",
                            "margin": "auto",
                        },
                    ),
                ]
            ),
            id={"type": "collapse", "index": 4},
            is_open=False,
        ),
        html.Div(children="", style={"fontSize": "24px", "marginBottom": "20px"}),
        dbc.Button(
            "Explore by feature",
            id={"type": "toggle-button", "index": 5},
            n_clicks=0,
            className="mb-3",
            disabled=True,
        ),
        dbc.Collapse(
            html.Div(
                [
                    html.Div(children="Timestamp", style={"marginBottom": "20px"}),
                    dcc.Dropdown(
                        id={"type": "timestamp-dropdown", "index": 6},
                        options=timestamp_options,
                        placeholder="Select timestamp",
                    ),
                    dcc.Dropdown(
                        id="features-dropdown",
                        options=feature_options,
                        placeholder="Select feature",
                    ),
                    daq.BooleanSwitch(
                        id="add-frequency-numbers-switch",
                        label="Add column with number of tests that have feature",
                        on=True
                    ),
                    html.Div(
                        [
                            dcc.Loading(
                                dash_table.DataTable(
                                    id="explore-by-feature-table",
                                    filter_action="native",
                                    sort_action="native",
                                    style_cell={
                                        "textAlign": "center",
                                        "minWidth": "100px",
                                        "maxWidth": "500px",
                                        "whiteSpace": "normal",
                                    },
                                    style_table={
                                        "overflowX": "auto",
                                        "maxWidth": "900px",
                                        "margin": "auto",
                                    },
                                ),
                            ),
                            html.Div(
                                id="explore-by-feature-data-container",
                                style={
                                    "display": "inline-block",
                                    "marginLeft": "20px",
                                    "verticalAlign": "top",
                                    "flex": 1,
                                    "overflow": "auto",
                                    "maxWidth": "900px",
                                },
                            ),
                        ],
                        id="explore-by-feature-container",
                        style={
                            "display": "flex",
                            "justifyContent": "center",
                            "alignItems": "flex-start",
                            "gap": "20px",
                            "maxWidth": "1900px",
                            "margin": "auto",
                        },
                    ),
                ]
            ),
            id={"type": "collapse", "index": 5},
            is_open=False,
        ),
    ]
)




if __name__ == "__main__":
    app.run(debug=False)



#!/usr/bin/env python3
"""Mock the /sys/class/gpio subsystem"""

import atexit
import logging
import os
import selectors
import shutil
import socket
import subprocess
import sys
import tempfile

from typing import Type


def read_gpio_pin(read_fd: int) -> str:
    """
    read_gpio_pin reads from the given fd and return a string with the
    numeric pin or "" if an invalid pin was specified.
    """
    data = os.read(read_fd, 128)
    pin = data.decode().strip()
    if not pin.isnumeric():
        logging.warning("invalid gpio pin %s", pin)
        return ""
    return "gpio{}".format(pin)


def export_ready(read_fd: int, _) -> bool:
    """export_ready is run when the "export" file is ready for reading"""
    pin = read_gpio_pin(read_fd)
    # allow quit
    if pin == "quit":
        return False
    if pin:
        with open(pin, "w"):
            pass
    return True


def unexport_ready(read_fd: int, _) -> bool:
    """export_ready is run when the "unexport" file is ready for reading"""
    pin = read_gpio_pin(read_fd)
    try:
        os.remove(pin)
    except OSError as err:
        logging.warning("got exception %s", err)
    return True


def dispatch(sel):
    """
    dispatch dispatches the events from the "sel" source
    """
    for key, mask in sel.select():
        callback = key.data
        if not callback(key.fileobj, mask):
            return False
    return True


def maybe_sd_notify(s: str) -> None:
    addr = os.getenv("NOTIFY_SOCKET")
    if not addr:
        return
    soc = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    soc.connect(addr)
    soc.sendall(s.encode())


def main():
    """the main method"""
    if os.getuid() != 0:
        print("must run as root")
        sys.exit(1)

    # setup mock env
    mock_gpio_dir = tempfile.mkdtemp("mock-gpio")
    atexit.register(shutil.rmtree, mock_gpio_dir)
    os.chdir(mock_gpio_dir)
    subprocess.check_call(["mount", "--bind", mock_gpio_dir, "/sys/class/gpio"])
    atexit.register(lambda: subprocess.call(["umount", "/sys/class/gpio"]))

    # fake gpio export/unexport files
    os.mkfifo("export")
    os.mkfifo("unexport")

    # react to the export/unexport calls
    sel = selectors.DefaultSelector()
    efd = os.open("export", os.O_RDWR | os.O_NONBLOCK)
    ufd = os.open("unexport", os.O_RDWR | os.O_NONBLOCK)
    sel.register(efd, selectors.EVENT_READ, export_ready)
    sel.register(ufd, selectors.EVENT_READ, unexport_ready)
    # notify
    maybe_sd_notify("READY=1")
    while True:
        if not dispatch(sel):
            break

    # cleanup when we get a quit call
    os.close(efd)
    os.close(ufd)


if __name__ == "__main__":
    main()
